/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "conntrack.h"
#include "vuurmuur.h"

char *
conn_get_status_name(const int debuglvl, int status)
{
	/* status needs to be smaller than the number of items in the array */
	if(status >= ( sizeof(conn_status) / sizeof(conn_status[0]) ) -1 )
		return(NULL);

	return(conn_status[status]);
}

//- print_dlist
void
conn_print_dlist(const d_list *dlist)
{
	d_list_node		*d_node = NULL;
	struct ConntrackData	*cd_ptr = NULL;
	char			status[16] = "";
	char			direction[16] = "";

	if(!dlist)
		return;

	for(d_node = dlist->top; d_node; d_node = d_node->next)
	{
		cd_ptr = d_node->data;

		if(cd_ptr->use_connect_status)
		{
			if(cd_ptr->connect_status == CONN_UNUSED)
				strcpy(status, "");
			else if(cd_ptr->connect_status == CONN_CONNECTING)
				strcpy(status, "CONNECTING");
			else if(cd_ptr->connect_status == CONN_CONNECTED)
				strcpy(status, "CONNECTED");
			else if(cd_ptr->connect_status == CONN_DISCONNECTING)
				strcpy(status, "DISCONNECTING");
			else
				strcpy(status, "UNKNOWN");
		}
		else
		{
			strcpy(status, "");
		}

		if(cd_ptr->direction_status == CONN_UNUSED)
			strcpy(direction, "");
		else if(cd_ptr->direction_status == CONN_IN)
			strcpy(direction, "INCOMING");
		else if(cd_ptr->direction_status == CONN_OUT)
			strcpy(direction, "OUTGOING");
		else if(cd_ptr->direction_status == CONN_FW)
			strcpy(direction, "FORWARDING");

		fprintf(stdout, "%4d: service %s from %s to %s %s %s\n", cd_ptr->cnt, cd_ptr->sername, cd_ptr->fromname, cd_ptr->toname, status, direction);
	}

	return;
}

/*	conntrack_line_to_data

	This function analyzes the line supplied through the connline_ptr. It should never fail,
	unless we have a serious problem: malloc failure or parameter problems.

	Returncodes:
		 0: ok
		-1: (serious) error
*/
int
conn_line_to_data(	const int debuglvl,
			struct ConntrackLine *connline_ptr,
			struct ConntrackData *conndata_ptr,
			Hash *serhash,
			Hash *zonehash,
			d_list *zonelist,
			int unknown_host_as_network,
			int sort_by_connect_status,
			int sort_by_in_out_fw
		)
{
	char		service_name[MAX_SERVICE],
			zone_name[MAX_HOST_NET_ZONE],
			*zone_name_ptr = NULL;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "conntrack_line_to_data called with: unknown_host_as_network: %d, sort_by_connect_status: %d.", unknown_host_as_network, sort_by_connect_status);

	/* safety */
	if(!connline_ptr || !conndata_ptr || !serhash || !zonehash)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: conntrack_line_to_data).");
		return(-1);
	}

	if(unknown_host_as_network && !zonelist)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem. No zonelist (in: conntrack_line_to_data).");
		return(-1);
	}

	/* first the service name */
	conndata_ptr->service = search_service_in_hash(debuglvl, connline_ptr->src_port, connline_ptr->dst_port, connline_ptr->protocol, serhash);
	if(conndata_ptr->service == NULL)
	{
		snprintf(service_name, sizeof(service_name), "%d -> %d", connline_ptr->src_port, connline_ptr->dst_port);
		if(!(conndata_ptr->sername = malloc(strlen(service_name)+1)))
		{
			(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: conntrack_line_to_data).", strerror(errno));
			return(-1);
		}

//TODO strlcpy
		strcpy(conndata_ptr->sername, service_name);
	}
	else
	{
		conndata_ptr->sername = conndata_ptr->service->name;
	}

	/* for hashing */
	conndata_ptr->dst_port = connline_ptr->dst_port;
	conndata_ptr->protocol = connline_ptr->protocol;
	/* for detailed data */
	conndata_ptr->src_port = connline_ptr->src_port;

	// then the from name
	conndata_ptr->from = search_zone_in_hash_with_ipv4(debuglvl, connline_ptr->src_ip, zonehash);
	if(conndata_ptr->from == NULL)
	{
		if(debuglvl >= HIGH) (void)vrprint.debug("Debug", "conntrack_line_to_data: we don't know the ip: '%s'.", connline_ptr->src_ip);

		if(!unknown_host_as_network)
		{
			snprintf(zone_name, sizeof(zone_name), "%s", connline_ptr->src_ip);

			if(!(conndata_ptr->fromname = malloc(strlen(zone_name)+1)))
			{
				(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: conntrack_line_to_data).", strerror(errno));
				return(-1);
			}
			else
				strcpy(conndata_ptr->fromname, zone_name);
		}
		else
		{
 			if(!(zone_name_ptr = get_network_for_ipv4(debuglvl, connline_ptr->src_ip, zonelist)))
			{
				if(!(conndata_ptr->fromname = malloc(strlen(connline_ptr->src_ip)+1)))
				{
					(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: conntrack_line_to_data).", strerror(errno));
					return(-1);
				}
				else
					strcpy(conndata_ptr->fromname, connline_ptr->src_ip);
			}
			else
			{
				if(!(conndata_ptr->fromname = malloc(strlen(zone_name_ptr)+1)))
				{
					(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: conntrack_line_to_data).", strerror(errno));
					return(-1);
				}
				else
					strcpy(conndata_ptr->fromname, zone_name_ptr);

				free(zone_name_ptr);
			}
		}
	}
	else
	{
		conndata_ptr->fromname = conndata_ptr->from->name;
	}

	// then the to name
	conndata_ptr->to = search_zone_in_hash_with_ipv4(debuglvl, connline_ptr->dst_ip, zonehash);
	if(conndata_ptr->to == NULL)
	{
		if(!unknown_host_as_network)
		{
			snprintf(zone_name, sizeof(zone_name), "%s", connline_ptr->dst_ip);
			if(!(conndata_ptr->toname = malloc(strlen(zone_name)+1)))
			{
				(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: conntrack_line_to_data).", strerror(errno));
				return(-1);
			}
			else
//TODO strlcpy
				strcpy(conndata_ptr->toname, zone_name);
		}
		else
		{
			if(!(zone_name_ptr = get_network_for_ipv4(debuglvl, connline_ptr->dst_ip, zonelist)))
			{
				if(!(conndata_ptr->toname = malloc(strlen(connline_ptr->dst_ip)+1)))
				{
					(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: conntrack_line_to_data).", strerror(errno));
					return(-1);
				}
				else
//TODO strlcpy
					strcpy(conndata_ptr->toname, connline_ptr->dst_ip);
			}
			else
			{
				if(!(conndata_ptr->toname = malloc(strlen(zone_name_ptr)+1)))
				{
					(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: conntrack_line_to_data).", strerror(errno));
					return(-1);
				}
				else
//TODO strlcpy
					strcpy(conndata_ptr->toname, zone_name_ptr);

				free(zone_name_ptr);
			}
		}
	}
	else
	{
		conndata_ptr->toname = conndata_ptr->to->name;
	}

	// status
	conndata_ptr->state = connline_ptr->state;

	// get status name. If it fails get the last in the array, which says (or at least should say) 'ERROR'
	if(!(conndata_ptr->status = conn_get_status_name(debuglvl, conndata_ptr->state)))
		conndata_ptr->status = conn_status[( sizeof(conn_status) / sizeof(conn_status[0]) )];

	conndata_ptr->use_connect_status = sort_by_connect_status;

	if(conndata_ptr->state == SYN_SENT || conndata_ptr->state == SYN_RECV || conndata_ptr->state == UNREPLIED)
		conndata_ptr->connect_status = CONN_CONNECTING;
	else if(conndata_ptr->state == TCP_ESTABLISHED || conndata_ptr->state == UDP_ESTABLISHED)
		conndata_ptr->connect_status = CONN_CONNECTED;
	else if(conndata_ptr->state == FIN_WAIT || conndata_ptr->state == TIME_WAIT || conndata_ptr->state == CLOSE || conndata_ptr->state == CLOSE_WAIT)
		conndata_ptr->connect_status = CONN_DISCONNECTING;
	else
		conndata_ptr->connect_status = CONN_UNUSED;

	if(conndata_ptr->from != NULL && conndata_ptr->from->type == TYPE_FIREWALL)
		conndata_ptr->direction_status = CONN_OUT;
	else if(conndata_ptr->to != NULL && conndata_ptr->to->type == TYPE_FIREWALL)
		conndata_ptr->direction_status = CONN_IN;
	else
		conndata_ptr->direction_status = CONN_FW;

	return(0);
}

// process one line
int
conn_process_one_conntrack_line(const int debuglvl, const char *line, struct ConntrackLine *connline_ptr)
{
	char	source_ip[32],
		dest_ip[32],

		source_port[16],
		dest_port[16],

		tmp_src_port[6],
		tmp_dst_port[6],

		protocol[16],
		status[16],

		id[16];

	// first determine protocol
	sscanf(line, "%s", protocol);

	if(strcmp(protocol, "tcp") == 0)
	{
		sscanf(line, "%16s %d %d %s %s %s %s %s", protocol, &connline_ptr->protocol, &connline_ptr->ttl, status, source_ip, dest_ip, source_port, dest_port);
	}
	else if(strcmp(protocol, "udp") == 0)
	{
		sscanf(line, "%16s %d %d %s %s %s %s %s", protocol, &connline_ptr->protocol, &connline_ptr->ttl, source_ip, dest_ip, source_port, dest_port, status);

		/* if the string starts with 'src' we have no status */
		if(strncmp(status, "src", 3) == 0)
			strcpy(status, "UDP_ESTABLISHED");
	}
	//icmp     1 29 src=192.168.0.2 dst=194.109.6.11 type=8 code=0 id=57376 [UNREPLIED] src=194.109.6.11 dst=192.168.0.2 type=0 code=0 id=57376 use=1
	else if(strcmp(protocol, "icmp") == 0)
	{
		sscanf(line, "%16s %d %d %s %s %s %s %s %s", protocol, &connline_ptr->protocol, &connline_ptr->ttl, source_ip, dest_ip, source_port, dest_port, id, status);
	}
	else if(strcmp(protocol, "unknown") == 0)
	{
		sscanf(line, "%16s %d %d %s %s %s %s %s", protocol, &connline_ptr->protocol, &connline_ptr->ttl, source_ip, dest_ip, source_port, dest_port, status);

		if(connline_ptr->protocol == 47 || connline_ptr->protocol == 50 || connline_ptr->protocol == 51)
		{
			strcpy(status, "none");
			strcpy(source_port, "");
			strcpy(dest_port, "");
		}
		else
		{
			strcpy(status, "none");
			strcpy(source_port, "sport=ERROR");
			strcpy(dest_port,   "dport=ERROR");
		}
	}
	else
	{
		strcpy(status, "none");
		connline_ptr->protocol = 0;
		strcpy(source_ip, "src=PARSE-ERROR");
		strcpy(dest_ip,   "dst=PARSE-ERROR");
		strcpy(source_port, "sport=ERROR");
		strcpy(dest_port,   "dport=ERROR");
	}

	/* src & dst ip */
	if(range_strcpy(connline_ptr->src_ip, source_ip, 4, strlen(source_ip), sizeof(connline_ptr->src_ip)) < 0)
		strcpy(connline_ptr->src_ip, "PARSE-ERROR");
	if(range_strcpy(connline_ptr->dst_ip, dest_ip, 4, strlen(dest_ip), sizeof(connline_ptr->dst_ip)) < 0)
		strcpy(connline_ptr->dst_ip, "PARSE-ERROR");

	/* ports - icmp has type= and code= both 5, while tcp and udp have sport= and dport= both 6 */
	if(connline_ptr->protocol == 1)
	{
		if(range_strcpy(tmp_src_port, source_port, 5, strlen(source_port), sizeof(tmp_src_port)) < 0)
			strcpy(tmp_src_port, "ERROR");
		if(range_strcpy(tmp_dst_port, dest_port, 5, strlen(dest_port), sizeof(tmp_dst_port)) < 0)
			strcpy(tmp_dst_port, "ERROR");
	}
	else
	{
		if(range_strcpy(tmp_src_port, source_port, 6, strlen(source_port), sizeof(tmp_src_port)) < 0)
			strcpy(tmp_src_port, "ERROR");
		if(range_strcpy(tmp_dst_port, dest_port, 6, strlen(dest_port), sizeof(tmp_dst_port)) < 0)
			strcpy(tmp_dst_port, "ERROR");
	}

	connline_ptr->src_port = atoi(tmp_src_port);
	if(connline_ptr->src_port <= 0 || connline_ptr->src_port > 65535)
		connline_ptr->src_port = 0;

	connline_ptr->dst_port = atoi(tmp_dst_port);
	if(connline_ptr->dst_port <= 0 || connline_ptr->dst_port > 65535)
		connline_ptr->dst_port = 0;

	/* process status */
	if(strcmp(status, "none") == 0)
		connline_ptr->state = NONE;
	else if(strcmp(status, "ESTABLISHED") == 0)
		connline_ptr->state = TCP_ESTABLISHED;
	else if(strcmp(status, "UDP_ESTABLISHED") == 0)
		connline_ptr->state = UDP_ESTABLISHED;
	else if(strcmp(status, "SYN_SENT") == 0)
		connline_ptr->state = SYN_SENT;
	else if(strcmp(status, "SYN_RECV") == 0)
		connline_ptr->state = SYN_RECV;
	else if(strcmp(status, "FIN_WAIT") == 0)
		connline_ptr->state = FIN_WAIT;
	else if(strcmp(status, "TIME_WAIT") == 0)
		connline_ptr->state = TIME_WAIT;
	else if(strcmp(status, "CLOSE") == 0)
		connline_ptr->state = CLOSE;
	else if(strcmp(status, "CLOSE_WAIT") == 0)
		connline_ptr->state = CLOSE_WAIT;
	else if(strcmp(status, "[UNREPLIED]") == 0)
		connline_ptr->state = UNREPLIED;
	else
		connline_ptr->state = UNDEFINED;

	return(0);
}

/*
	type:
		1. service
		2. from zone
		3. to zone
*/
int
conn_add_service_hostip_to_list_and_hash(	const int debuglvl,
						d_list *list,
						Hash *hash,
						struct ConntrackData *conndata_ptr,
						int type
					)
{
	struct ConntrackItem_	*item_ptr=NULL,
				*old_item_ptr=NULL;


	/* safety */
	if(!list || !hash || !conndata_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* alloc conn item */
	if(!(item_ptr = malloc(sizeof(struct ConntrackItem_))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/* temp store */
	old_item_ptr = item_ptr;

	
	/* service */
	if(type == 1)
	{
		/* set values */
		if(conndata_ptr->service == NULL)
			snprintf(item_ptr->name, sizeof(item_ptr->name), "unknown(%d(%d))", conndata_ptr->dst_port, conndata_ptr->protocol);
		else
			snprintf(item_ptr->name, sizeof(item_ptr->name), "%s", conndata_ptr->sername);
	}
	/* from */
	else if(type == 2)
	{
		strlcpy(item_ptr->name, conndata_ptr->fromname, sizeof(item_ptr->name));
	}
	/* to */
	else if(type == 3)
	{
		strlcpy(item_ptr->name, conndata_ptr->toname, sizeof(item_ptr->name));
	}


	/* first check if we already have it */
	if(hash_search(debuglvl, hash, (void **)&item_ptr) == 0)
	{
		/* found, we already have it in memory */
		free(old_item_ptr);

		/* increase counter */
		item_ptr->count++;
	}
	else
	{
		/*	prepend into the list
		*/
		if(d_list_prepend(debuglvl, list, item_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "inserting into list failed (in: %s:%d).", __FUNCTION__, __LINE__);

			free(item_ptr);
			return(-1);
		}

		if(hash_insert(debuglvl, hash, item_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "inserting into hash failed (in: %s:%d).", __FUNCTION__, __LINE__);

			free(item_ptr);
			return(-1);
		}

		item_ptr->count = 1;
	}

	return(0);
}


/*	conn_hash_name

	Very simple string hashing function. It just adds up
	all chars.
*/
unsigned int
conn_hash_name(const void *key)
{
	int		len = 0;
	unsigned int	hash = 0;
	char		*name = NULL;

	if(!key)
		return(1);

	name = (char *)key;

	len = strlen(name);
	while(len)
	{
		hash = hash + name[len];
		len--;
	}

	return(hash);
}


//TODO silly names
int
conn_match_name(const void *ser1, const void *ser2)
{
	if(!ser1 || !ser2)
		return(0);

	if(strcmp((char *)ser1, (char *)ser2) == 0)
		return 1;
	else
		return 0;
}

//- print_list -
void
conn_list_print(const d_list *conn_list)
{
	d_list_node		*d_node;
	struct ConntrackItem_	*seritem_ptr;

	// Display the linked list.
	fprintf(stdout, "List len is %d\n", conn_list->len);

	for(d_node = conn_list->top; d_node ; d_node = d_node->next)
	{
		seritem_ptr = d_node->data;

		fprintf(stdout, "name: %s, count: %d\n", seritem_ptr->name, seritem_ptr->count);
	}

	return;
}


unsigned int
conn_hash_string(const void *key)
{
	const char	*ptr;
	unsigned int	val = 0;

	ptr = key;

	while(*ptr != '\0')
	{
		int tmp;

		val = (val << 4) + (*ptr);

		if((tmp = (val & 0xf0000000)))
		{
			val = val ^ (tmp >> 24);
			val = val ^ tmp;
		}
		ptr++;
	}

	return(val);
}


/*	hash_conntrackdata

	Hashes conntrackdata. It does this by creating seperate
	hashes for sername, fromname and toname.

	Returns the hash.
*/
unsigned int
conn_hash_conntrackdata(const void *key)
{
	unsigned int		retval = 0;
	struct ConntrackData	*cd_ptr;

	if(!key)
		return(1);

	cd_ptr = (struct ConntrackData *)key;

	/*	from and to have different weight, so firewall -> internet
		is not the same as internet -> firewall
	*/
	retval = retval + conn_hash_name(cd_ptr->sername);
	retval = retval + conn_hash_name(cd_ptr->fromname) / 2;
	retval = retval + conn_hash_name(cd_ptr->toname) / 3;

	if(cd_ptr->use_connect_status)
	{
		// hash with the connection status
		if(cd_ptr->connect_status == CONN_CONNECTING)
			retval=retval/4;
		else if(cd_ptr->connect_status == CONN_CONNECTED)
			retval=retval/5;
		else if(cd_ptr->connect_status == CONN_DISCONNECTING)
			retval=retval/6;
	}

	return(retval);
}


/*	match_conntrackdata

*/
int
conn_match_conntrackdata(const void *check, const void *hash)
{
	struct ConntrackData	*check_cd = NULL,
				*hash_cd = NULL;

	/* safety */
	if(!check || !hash)
		return(0);

	check_cd = (struct ConntrackData *)check;
	hash_cd  = (struct ConntrackData *)hash;
	if(!check_cd || !hash_cd)
	{
		(void)vrprint.error(0, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
		return(0);
	}

	if(strncmp(check_cd->sername, hash_cd->sername, MAX_SERVICE) == 0)
	{
		// service matches
		if(strncmp(check_cd->fromname, hash_cd->fromname, MAX_HOST_NET_ZONE) == 0)
		{
			// from host also matches
			if(strncmp(check_cd->toname, hash_cd->toname, MAX_HOST_NET_ZONE) == 0)
			{
				// if we don't use this, they are both 0, so they match
				if(check_cd->use_connect_status)
				{
					if(check_cd->connect_status == hash_cd->connect_status)
					{
						// they all match-> return 1
						return(1);
					}
				}
				else
				{
					return(1);
				}
			}
		}
	}

	// sorry, no match
	return(0);
}


/*	conn_dlist_destroy

	Destroys the list.
*/
void conn_list_cleanup(int debuglvl, d_list *conn_dlist)
{
	d_list_node		*d_node = NULL;
	struct ConntrackData	*cd_ptr = NULL;

	for(d_node = conn_dlist->top; d_node; d_node = d_node->next)
	{
		cd_ptr = d_node->data;

		if(cd_ptr->from == NULL)
			free(cd_ptr->fromname);
		if(cd_ptr->to == NULL)
			free(cd_ptr->toname);
		if(cd_ptr->service == NULL)
			free(cd_ptr->sername);

		free(cd_ptr);
	}

	d_list_cleanup(debuglvl, conn_dlist);
}


/*	conn_get_connections

	Assembles all conntrack connections in one list, and counts all items.

	prev_conn_cnt is used to determine the size of the hashtable which is used. It is based on
	the size of the list of the last time we ran this function. If it is zero, we use a default.

	TODO:	intergrate with get stats
		group results on:	network when unknown host - done
					fw, in, out
					connected, connecting, disconnecting

		make sorting better by check if current cd is bigger than the top 3

	Do this by the way we create a hash, so set the options into the cd struct
*/
int
conn_get_connections(	const int debuglvl,
			const int prev_conn_cnt,
			Hash *serv_hash,
			Hash *zone_hash,
			d_list *conn_dlist,
			d_list *zone_list,
			const int unknown_host_as_network,
			const int sort_by_connect_status,
			const int sort_by_in_out_fw,
			struct ConntrackStats_ *connstat_ptr,
			/*@null@*/char *zonefilter,
			/*@null@*/char *servicefilter
		)
{
	int			retval = 0;

	char			line[256];
	FILE			*fp = NULL;
	struct ConntrackLine	cl;
	struct ConntrackData	*cd_ptr = NULL,
				*old_cd_ptr = NULL,
				*prev_cd_ptr = NULL,
				*next_cd_ptr = NULL;

	/* default hashtable size */
	int			hashtbl_size = 256;
	Hash			conn_hash;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!serv_hash || !zone_hash || prev_conn_cnt < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* if the prev_conn_cnt supplied by the user is bigger than 0,
	   use it. */
	if(prev_conn_cnt > 0)
		hashtbl_size = prev_conn_cnt;

	/* initialize the hash */
	if(hash_setup(debuglvl, &conn_hash, hashtbl_size, conn_hash_conntrackdata, conn_match_conntrackdata) != 0)
	{
		(void)vrprint.error(-1, "Internal Error", "unable to initialize hash (in: conn_get_connections).");
		return(-1);
	}

	/* open conntrack file (fopen) */
	if(!(fp = fopen(PROC_IPCONNTRACK, "r")))
	{
		(void)vrprint.error(-1, "Error", "unable to open %s: %s (in: conn_get_connections).", PROC_IPCONNTRACK, strerror(errno));
		return(-1);
	}

	/* set stat counters to zero */
	connstat_ptr->conn_total = 0,
	connstat_ptr->conn_in = 0,
	connstat_ptr->conn_out = 0,
	connstat_ptr->conn_fw = 0;

	connstat_ptr->stat_connect = 0,
	connstat_ptr->stat_estab = 0,
	connstat_ptr->stat_closing = 0,
	connstat_ptr->stat_other = 0;


	/*	now read the file, interpret the line and trough hash_look up if the line is already in the list

		if it is do 2 things:
			1. increment the counter
			2. check if the count is bigger than the line above (in the list)
				if so, move line up one

		else insert the line into the list, and hash

		The result will be reasonably good sorted list, at almost no speed penalty.
	*/
	while((fgets(line, sizeof(line), fp) != NULL))
	{
		/* parse the line */
		if(conn_process_one_conntrack_line(debuglvl, line, &cl) < 0)
		{
			(void)vrprint.error(-1, "Error", "parsing conntrack line '%s' failed (in: conn_get_connections).", line);
			return(-1);
		}

		/* allocate memory for the data */
		if(!(cd_ptr = (struct ConntrackData *)malloc(sizeof(struct ConntrackData))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed, out of memory (in: conn_get_connections).");
			return(-1);
		}

		/* analyse it */
		if(conn_line_to_data(debuglvl, &cl, cd_ptr, serv_hash, zone_hash, zone_list, unknown_host_as_network, sort_by_connect_status, sort_by_in_out_fw) < 0)
		{
			(void)vrprint.error(-1, "Error", "analizing conntrack line failed (in: conn_get_connections).");

			free(cd_ptr);
			return(-1);
		}

		/* if the hashlookup is succesfull, cd_ptr is overwritten, so we store it here */
		old_cd_ptr = cd_ptr;

		/* we ignore the local loopback  connections */
		if((strncmp(cd_ptr->fromname, "127.", 4) == 0 || strncmp(cd_ptr->toname,   "127.", 4) == 0))
		{
			if(cd_ptr->from == NULL)
				free(cd_ptr->fromname);
			if(cd_ptr->to == NULL)
				free(cd_ptr->toname);
			if(cd_ptr->service == NULL)
				free(cd_ptr->sername);

			free(cd_ptr);
			cd_ptr = NULL;
			old_cd_ptr = NULL;
		}
		else
		{
			/* update counters */
			connstat_ptr->conn_total++;

			if(cd_ptr->from != NULL && cd_ptr->from->type == TYPE_FIREWALL)
				connstat_ptr->conn_out++;
			else if(cd_ptr->to != NULL && cd_ptr->to->type == TYPE_FIREWALL)
				connstat_ptr->conn_in++;
			else
				connstat_ptr->conn_fw++;

			if(cd_ptr->connect_status == CONN_CONNECTING)
				connstat_ptr->stat_connect++;
			else if(cd_ptr->connect_status == CONN_DISCONNECTING)
				connstat_ptr->stat_closing++;
			else if(cd_ptr->connect_status == CONN_CONNECTED)
				connstat_ptr->stat_estab++;
			else
				connstat_ptr->stat_other++;

			/* now check if the cd is already in the list */
			if((cd_ptr = hash_search(debuglvl, &conn_hash, (void *)cd_ptr)) != NULL)
			{
				/*	FOUND in the hash

					free the memory in the old_cd_ptr, we dont need it no more */
				if(old_cd_ptr->from == NULL)
					free(old_cd_ptr->fromname);
				if(old_cd_ptr->to == NULL)
					free(old_cd_ptr->toname);
				if(old_cd_ptr->service == NULL)
					free(old_cd_ptr->sername);

				free(old_cd_ptr);
				old_cd_ptr = NULL;

				/* now increment the counter */
				cd_ptr->cnt++;

				/* check if the above cd in the list is smaller than we are */
				if((d_node = cd_ptr->d_node->prev))
				{
					prev_cd_ptr = d_node->data;

					if(cd_ptr->cnt > prev_cd_ptr->cnt)
					{
						/* yes, so now we move one up */
						if(d_list_remove_node(debuglvl, conn_dlist, cd_ptr->d_node) < 0)
						{
							(void)vrprint.error(-1, "Internal Error", "removing from list failed (in: conn_get_connections).");
							return(-1);
						}

						/* now reinsert */
						if(!(cd_ptr->d_node = d_list_insert_before(debuglvl, conn_dlist, d_node, cd_ptr)))
						{
							(void)vrprint.error(-1, "Internal Error", "unable to insert into list (in: conn_get_connections).");
							return(-1);
						}
					}
					/*	check if the beneath cd in the list is bigger than we are,
						 we only do this if the above wasn't smaller
					*/
					else if((d_node = cd_ptr->d_node->next))
					{
						next_cd_ptr = d_node->data;

						if(cd_ptr->cnt < next_cd_ptr->cnt)
						{
							/* yes, so now we move one down */
							if(d_list_remove_node(debuglvl, conn_dlist, cd_ptr->d_node) < 0)
							{
								(void)vrprint.error(-1, "Internal Error", "removing from list failed (in: conn_get_connections).");
								return(-1);
							}

							/* now reinsert */
							if(!(cd_ptr->d_node = d_list_insert_after(debuglvl, conn_dlist, d_node, cd_ptr)))
							{
								(void)vrprint.error(-1, "Internal Error", "unable to insert into list (in: conn_get_connections).");
								return(-1);
							}
						}
					}
				}

				/*
					now we do one last check
				*/

				/* check if the one above us is 1, if so, move to bottom of the list */
				if((d_node = cd_ptr->d_node->prev))
				{
					prev_cd_ptr = d_node->data;

					/*	is the one beneath us is 1 and not the bottom of the list,
						move it to the bottom of the list */
					if(prev_cd_ptr->cnt == 1 && d_node->prev != NULL)
					{
						/* yes, so now we first remove */
						if(d_list_remove_node(debuglvl, conn_dlist, d_node) < 0)
						{
							(void)vrprint.error(-1, "Internal Error", "removing from list failed (in: conn_get_connections).");
							return(-1);
						}

						/* and then re-insert */
						if(!(prev_cd_ptr->d_node = d_list_append(debuglvl, conn_dlist, prev_cd_ptr)))
						{
							(void)vrprint.error(-1, "Internal Error", "unable to insert into list (in: conn_get_connections).");
							return(-1);
						}
					}
				}

				/* do the same for the one below us */
				if((d_node = cd_ptr->d_node->next))
				{
					next_cd_ptr = d_node->data;

					/* is the one beneath us is 1 and not the bottom of the list, 
					   move it to the bottom */
					if(next_cd_ptr->cnt == 1 && d_node->next != NULL)
					{
						/* yes, so now remove */
						if(d_list_remove_node(debuglvl, conn_dlist, d_node) < 0)
						{
							(void)vrprint.error(-1, "Internal Error", "removing from list failed (in: conn_get_connections).");
							return(-1);
						}

						/* now reinsert */
						if(!(next_cd_ptr->d_node = d_list_append(debuglvl, conn_dlist, next_cd_ptr)))
						{
							(void)vrprint.error(-1, "Internal Error", "unable to insert into list (in: conn_get_connections).");
							return(-1);
						}
					}
				}
			}
			else
			{
				/*	NOT found in the hash

					set cd_ptr to old_cd_ptr because cd_ptr is NULL after the failed hash search
				*/
				cd_ptr = old_cd_ptr;

				/* append the new cd to the list */
				cd_ptr->d_node = d_list_append(debuglvl, conn_dlist, cd_ptr);
				if(!cd_ptr->d_node)
				{
					(void)vrprint.error(-1, "Internal Error", "unable to append into list (in: conn_get_connections).");
					return(-1);
				}

				/* and insert it into the hash */
				if(hash_insert(debuglvl, &conn_hash, cd_ptr) != 0)
				{
					(void)vrprint.error(-1, "Internal Error", "unable to insert into hash (in: conn_get_connections).");
					return(-1);
				}

				/* set cnt to 1 */
				cd_ptr->cnt = 1;
			}
		}
	}

	/* close the file */
	if(fclose(fp) < 0)
		retval = -1;

	/* cleanup */
	hash_cleanup(debuglvl, &conn_hash);

	return(retval);
}
