/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "vuurmuur.h"

/* - determine_action -
 * In this function we translate the 'accept' or 'drop' from the 'rules.conf' file to the
 * values that iptables understands, like 'ACCEPT, DROP, REJECT'.
 *
 * The function is called with the action from the rulesfile 'query' and returns the iptables
 * action trough 'action'.
 *
 * Returncodes:
 * 	0: ok, found
 *     -1: invalid query
 */
int
determine_action(const int debuglvl, char *query, char *action, size_t size, struct options *option)
{
	int retval=-1;

	if(!query || !action || !option)
		return(-1);

	if (debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "determine_action: ** start **");

	if((strncasecmp(query, "allow", 5) == 0) || (strncasecmp(query, "accept", 6) == 0))
	{
		strlcpy(action, "ACCEPT", size);
		retval=0;
	}

	if((strncasecmp(query, "deny", 4) == 0) || (strncasecmp(query, "drop", 4) == 0))
	{
		strlcpy(action, "DROP", size);
		retval=0;
	}

	if(strncasecmp(query, "reject", 6) == 0)
	{
		strlcpy(action, "REJECT", size);
		if(option->reject_option == 1)
		{
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug("Debug", "determine_action: reject option: '%s'.", option->reject_type);

			snprintf(action, size, "REJECT --reject-with %s", option->reject_type);
		}
		retval=0;
	}

	if(strncasecmp(query, "redirect", 8) == 0)
	{
		strlcpy(action, "REDIRECT", size);
		if(option->redirectport > 0)
		{
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug("Debug", "determine_action: redirect option: '%d'.", option->redirectport);

			snprintf(action, size, "REDIRECT --to-ports %d", option->redirectport);
			retval=0;
		}
		else
		{
			(void)vrprint.warning("Warning", "Target REDIRECT requires option 'redirectport'.");
		}
	}

	if(strncasecmp(query, "log", 3) == 0)
	{
		if(strcmp(conf.loglevel, "") == 0)
			strlcpy(action, "LOG", size);
		else
			snprintf(action, size, "LOG --log-level %s", conf.loglevel);

		retval=0;
	}

	if(strncasecmp(query, "masq", 4) == 0)
	{
		strlcpy(action, "MASQUERADE", size);
		retval=0;
	}

	if(strncasecmp(query, "snat", 4) == 0)
	{
		strlcpy(action, "SNAT", size);
		retval=0;
	}

	if((strncasecmp(query, "portfw", 6) == 0) || (strncasecmp(query, "dnat", 4) == 0))
	{
		strlcpy(action, "DNAT", size);
		retval=0;
	}

	if(strncasecmp(query, "queue", 5) == 0)
	{
		strlcpy(action, "QUEUE", size);
		retval=0;
	}

	if(debuglvl >= LOW)
	{
		if(retval==0)
		{
			(void)vrprint.debug("Debug", "determine_action: query = %s, this means: action = %s", query, action);
		}
	}

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "determine_action: ** end **, return=%d", retval);

	return(retval);
}


/*	determine_chain

	This function gets its info from the 'query' structure, and outputs the action string and
	the int ruletype.

	Returncodes:
		 0: ok
		-1: error
*/
int
determine_chain(const int debuglvl, struct RuleData_ *rule_ptr, char *chain, size_t size, int *ruletype)
{
	int	retval = -1;

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "determine_chain: ** start **");

	if(strncasecmp(rule_ptr->from, "firewall", 8) == 0)
	{
		//output
		strlcpy(chain, "OUTPUT", size);

		*ruletype = RT_OUTPUT;
		// fprintf(stdout, "determine_chain: output\n");
		retval=0;
	}

	if(strncasecmp(rule_ptr->to, "firewall", 8) == 0)
	{
		//input
		strlcpy(chain, "INPUT", size);

		*ruletype = RT_INPUT;
		// fprintf(stdout, "determine_chain: input\n");
		retval=0;
	}

	if((strncasecmp(rule_ptr->to, "firewall", 8) != 0) && (strncasecmp(rule_ptr->from, "firewall", 8) != 0))
	{
		// forward
		strlcpy(chain, "FORWARD", size);
		
		*ruletype = RT_FORWARD;
		retval=0;
	}

	if (strncasecmp(rule_ptr->action, "masq", 4) == 0)
	{
		strlcpy(chain, "POSTROUTING", size);
		*ruletype = RT_MASQ;
		retval=0;
	}

	if (strncasecmp(rule_ptr->action, "snat", 4) == 0)
	{
		strlcpy(chain, "POSTROUTING", size);
		*ruletype = RT_SNAT;
		retval=0;
	}

	/* prerouting chain for portfw/dnat */
	if((strncasecmp(rule_ptr->action, "portfw", 6) == 0) || (strncasecmp(rule_ptr->action, "dnat", 4) == 0))
	{
		strlcpy(chain, "PREROUTING", size);

		*ruletype = RT_PORTFW;

		retval = 0;
	}

	/* prerouting for redirect */
	if(strncasecmp(rule_ptr->action, "redirect", 8) == 0)
	{
		strlcpy(chain, "PREROUTING", size);

		*ruletype = RT_REDIRECT;

		retval = 0;
	}

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "determine_chain: ** end **, return=%d", retval);

	return(retval);
}

int
get_ip_info(const int debuglvl, char *name, struct ZoneData_ *answer_ptr, struct rgx_ *reg)
{
	int	retval = 0,
		result = 0;

	/* safety */
	if(!name || !answer_ptr || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "get_ip_info: ** start **");

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "get_ip_info: determining info for '%s'.", name);

	switch(answer_ptr->type)
	{
		case TYPE_HOST:

			/*
				ask the ipaddress for this host
			*/
			result = zf->ask(debuglvl, zone_backend, name, "IPADDRESS", answer_ptr->ipv4.ipaddress, sizeof(answer_ptr->ipv4.ipaddress), TYPE_HOST, 0);
			if(result == 1)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "get_ip_info: host %s has ipaddress %s.", name, answer_ptr->ipv4.ipaddress);
			}
			else if(result == 0)
			{
				(void)vrprint.warning("Warning", "No ipaddress defined for '%s', deactivating host.", name);
				answer_ptr->active = 0;
			}
			else
			{
				(void)vrprint.error(-1, "Error", "backend returned error when trying to get ipaddress for host '%s' (in: get_ip_info).", name);
				retval=-1;
			}

			/*
				get the mac-address
			*/
			answer_ptr->has_mac = get_mac_address(debuglvl, name, answer_ptr->mac, sizeof(answer_ptr->mac), reg->macaddr);
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "get_ip_info: has_mac: %d", answer_ptr->has_mac);

			/*
				we need the network and netmask of the network as well, so we can check the ip
			*/
			strlcpy(answer_ptr->ipv4.network, answer_ptr->network_parent->ipv4.network, sizeof(answer_ptr->ipv4.network));
			strlcpy(answer_ptr->ipv4.netmask, answer_ptr->network_parent->ipv4.netmask, sizeof(answer_ptr->ipv4.netmask));
			strlcpy(answer_ptr->ipv4.broadcast, answer_ptr->network_parent->ipv4.broadcast, sizeof(answer_ptr->ipv4.broadcast));

			/*
				if the network or netmask are not okay we already know that from the network, else check the ip
			*/
			if(strcmp(answer_ptr->ipv4.network, "") != 0 && strcmp(answer_ptr->ipv4.netmask, "") != 0 && strcmp(answer_ptr->ipv4.ipaddress, "") != 0)
			{
				/*
					check the ip
				*/
				result = check_ipv4address(debuglvl, answer_ptr->ipv4.network, answer_ptr->ipv4.netmask, answer_ptr->ipv4.ipaddress, 0);
				if(result < 0)
				{
					(void)vrprint.error(-1, "Error", "checking the ipaddress of '%s' failed (in: %s).", name, __FUNCTION__);
					retval = -1;
				}
				else if(result == 0)
				{
					/*
						check ip told us that the ip didn't belong to the network
					*/
					(void)vrprint.warning("Warning", "Ipaddress '%s' of host '%s' does not belong to network '%s' with netmask '%s'. Deactivating host.", answer_ptr->ipv4.ipaddress, name, answer_ptr->ipv4.network, answer_ptr->ipv4.netmask);
					answer_ptr->active = 0;
				}
			}

			/*
				for iptables, the netmask of a single host is always 255.255.255.255,
				we do this after check_ip because otherwise check_ip would not work
			*/
			strcpy(answer_ptr->ipv4.netmask, "255.255.255.255");

			break;

		case TYPE_NETWORK:

			/*
				get the network address
			*/
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "get_ip_info: get network_ip for '%s', max_size: %d.", name, sizeof(answer_ptr->ipv4.network));

			result = zf->ask(debuglvl, zone_backend, name, "NETWORK", answer_ptr->ipv4.network, sizeof(answer_ptr->ipv4.network), TYPE_NETWORK, 0);
			if(result == 1)
			{
				if(debuglvl >= MEDIUM) (void)vrprint.debug("Debug", "get_ip_info: network %s has network %s.", name, answer_ptr->ipv4.network);
			}
			else if(result == 0)
			{
				(void)vrprint.warning("Warning", "No network defined for '%s', deactivating network.", name);
				answer_ptr->active = 0;
			}
			else
			{
				(void)vrprint.error(-1, "Error", "get_ip_info: ask_backend returned error.");
				retval=-1;
			}

			/*
				netmask
			*/
			result = zf->ask(debuglvl, zone_backend, name, "NETMASK", answer_ptr->ipv4.netmask, sizeof(answer_ptr->ipv4.netmask), TYPE_NETWORK, 0);
			if(result == 1)
			{
				if(debuglvl >= MEDIUM) (void)vrprint.debug("Debug", "get_ip_info: network %s has netmask %s.", name, answer_ptr->ipv4.netmask);
			}
			else if(result == 0)
			{
				(void)vrprint.warning("Warning", "No netmask defined for '%s', deactivating network.", name);
				answer_ptr->active = 0;
			}
			else
			{
				(void)vrprint.error(-1, "Error", "get_ip_info: ask_backend returned error.");
				retval=-1;
			}

			/*
				get the broadcast address for this network/netmask combination
			*/
			if(strcmp(answer_ptr->ipv4.network, "") != 0 && strcmp(answer_ptr->ipv4.netmask, "") != 0)
			{
				result = create_broadcast_ip(debuglvl, answer_ptr->ipv4.network, answer_ptr->ipv4.netmask, answer_ptr->ipv4.broadcast, sizeof(answer_ptr->ipv4.broadcast));
				if(result != 0)
				{
					// error
					(void)vrprint.error(-1, "Error", "create_broadcast_ip for zonedata '%s' failed.", answer_ptr->name);
					answer_ptr->active = 0;
				}
			}
			break;

		default:
			(void)vrprint.error(-1, "Internal Error", "Expected a host or a network, got a %d (in: get_ip_info).", answer_ptr->type);
			retval=-1;
			break;
	}

	if (debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "get_ip_info: ** end **, return=%d", retval);

	return(retval);
}


/* split_and_check

    This function splits up the 'query.from' and 'query.to' values
    into host, network and zone data. All values are checked if they
    are active.

    Returncodes:
    -1: error
     0: ok
     1: some value not active
 */
int
split_and_check(const int debuglvl, Zones *zones, char *name, struct ZoneData_ *checkzone_ptr, regex_t *reg_ex)
{
	int	retval = 0,
		arg_count = 0,
		result = 0;
	char	check_str[MAX_HOST_NET_ZONE];
	char	zonename[MAX_ZONE],
		netname[MAX_NETWORK],
		hostname[MAX_HOST];

	/* safety */
	if(!name || !checkzone_ptr || !zones || !reg_ex)
	{
		(void)vrprint.error(-1, "Interal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "start: name = '%s'", name);

	/* validate and split up */
	if(validate_zonename(debuglvl, name, 0, zonename, netname, hostname, reg_ex) != 0)
	{
		(void)vrprint.error(-1, "Error", "zonename '%s' not valid.", name);
		return(-1);
	}
	else
	{
		arg_count = 2;

		if(hostname[0] == '\0')
			arg_count = 1;
		if(netname[0] == '\0')
			arg_count = 0;
		if(zonename[0] == '\0')
			return(-1);
	}

	/* zone or firewall */
	if(arg_count == 0 )
	{
		strlcpy(checkzone_ptr->zone_name, zonename, sizeof(checkzone_ptr->zone_name));

		result = check_active(debuglvl, checkzone_ptr->zone_name, TYPE_ZONE);
		if(result == 0)
		{
			retval=1;
		}
		else if(result == -1)
		{
			(void)vrprint.error(-1, "Error", "check_active returned an error while checking '%s'.", checkzone_ptr->zone_name);
			retval=-1;
		}

		if(debuglvl >= MEDIUM)
			(void)vrprint.debug(__FUNCTION__, "checkzone_ptr->zone = '%s'.", checkzone_ptr->zone_name);
	}

	/* network */
	if(arg_count == 1)
	{
		strlcpy(checkzone_ptr->network_name, netname, sizeof(checkzone_ptr->network_name));
		strlcpy(checkzone_ptr->zone_name, zonename, sizeof(checkzone_ptr->zone_name));

		checkzone_ptr->zone_parent = search_zonedata(debuglvl, zones, checkzone_ptr->zone_name);
		if(checkzone_ptr->zone_parent == NULL)
		{
			(void)vrprint.error(-1, "Error", "unable to find the zone '%s' in memory.", checkzone_ptr->zone_name);
			return(-1);
		}
		else
		{
			if(checkzone_ptr->zone_parent->active == 0)
			{
				retval = 1;
			}
		}

		snprintf(check_str, sizeof(check_str), "%s.%s", checkzone_ptr->network_name, checkzone_ptr->zone_name);
		result = check_active(debuglvl, check_str, TYPE_NETWORK);
		if(result == 0)
		{
			retval = 1;
		}
		else if(result == -1)
		{
			(void)vrprint.error(-1, "Error", "check_active returned an error while checking '%s'.", check_str);
			retval=-1;
		}

		if(debuglvl >= MEDIUM)
			(void)vrprint.debug(__FUNCTION__, "checkzone_ptr->network = '%s', checkzone_ptr->zone = '%s'.",
								checkzone_ptr->network_name,
								checkzone_ptr->zone_name);
	}

	/* host or group */
	if(arg_count == 2)
	{
		strlcpy(checkzone_ptr->host_name, hostname, sizeof(checkzone_ptr->host_name));
		strlcpy(checkzone_ptr->network_name, netname, sizeof(checkzone_ptr->network_name));
		strlcpy(checkzone_ptr->zone_name, zonename, sizeof(checkzone_ptr->zone_name));

		checkzone_ptr->zone_parent = search_zonedata(debuglvl, zones, checkzone_ptr->zone_name);
		if(checkzone_ptr->zone_parent == NULL)
		{
			(void)vrprint.error(-1, "Error", "unable to find the zone '%s' in memory.", checkzone_ptr->zone_name);
			return(-1);
		}
		else
		{
			if(checkzone_ptr->zone_parent->active == 0)
			{
				retval = 1;
			}
		}

		snprintf(check_str, sizeof(check_str), "%s.%s", checkzone_ptr->network_name, checkzone_ptr->zone_name);
		checkzone_ptr->network_parent = search_zonedata(debuglvl, zones, check_str);
		if(checkzone_ptr->network_parent == NULL)
		{
			(void)vrprint.error(-1, "Error", "Unable to find the network '%s' in memory.", check_str);
			retval=-1;
		}
		else
		{
			if(checkzone_ptr->network_parent->active == 0)
			{
				retval = 1;
			}
		}

		snprintf(check_str, sizeof(check_str), "%s.%s.%s", checkzone_ptr->host_name, checkzone_ptr->network_name, checkzone_ptr->zone_name);
		result=check_active(debuglvl, check_str, checkzone_ptr->type);
		if(result == 0)
		{
			retval=1;
		}
		else if(result == -1)
		{
			(void)vrprint.error(-1, "Error", "check_active returned an error while checking '%s'.", checkzone_ptr->zone_name);
			retval=-1;
		}

		if(debuglvl >= MEDIUM)
			(void)vrprint.debug(__FUNCTION__, "check->host = '%s', check->network = '%s', check->zone = '%s'.",
								checkzone_ptr->host_name,
								checkzone_ptr->network_name,
								checkzone_ptr->zone_name);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "** end **, return=%d", retval);

	return(retval);
}


/* create_broadcast_ip

    For broadcasting protocols we need an ipaddress to broadcast to.
    This function creates this ipaddress.

    Returncodes:
     0: ok
    -1: error
*/
int
create_broadcast_ip(const int debuglvl, char *network, char *netmask, char *broadcast_ip, size_t size)
{
	int retval=0;

	struct in_addr net;     /* the network address against we want to check */
	struct in_addr mask;    /* the netmask of the network */
	struct in_addr broad;   /* the broadcast address of this network */

	unsigned long int netmaskvalue=0;
	unsigned long int networkvalue=0;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "create_broadcast_ip: ** start **");

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "create_broadcast_ip: debuglvl: %d, network: %s, netmask: %s", debuglvl, network, netmask);

	if(inet_aton(netmask, &mask) == 0)
	{
		(void)vrprint.error(-1, "Error", "Invalid netmask: '%s' (in: create_broadcast_ip).", netmask);
		return(-1);
	}
	else
	{
		netmaskvalue=ntohl(mask.s_addr);

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "create_broadcast_ip: netmask = %s", inet_ntoa(mask));
	}

	if(inet_aton(network, &net) == 0)
	{
		(void)vrprint.error(-1, "Error", "Invalid network: '%s' (in: create_broadcast_ip).", network);
		return(-1);
	}
	else
	{
		networkvalue=ntohl(net.s_addr);

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "create_broadcast_ip: network = %s", inet_ntoa(net));
	}

	broad=net;
	broad.s_addr|=~ntohl(netmaskvalue);

	strlcpy(broadcast_ip, inet_ntoa(broad), size);

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "create_broadcast_ip: broadcast-address for network %s with netmask %s is %s.", network, netmask, broadcast_ip);

	if(debuglvl>=HIGH)
		(void)vrprint.debug("Debug", "create_broadcast_ip: ** end **, return=%d", retval);

	return(retval);
}


/*	get_group_info

	This function reads the groupfile.

	Returncodes:
		 0: ok
		-1: error
		
	TODO: rename 'query' var to something I understand.
	same for answer_ptr and d_list
 */
int
get_group_info(const int debuglvl, Zones *zones, char *groupname, struct ZoneData_ *answer_ptr)
{
	int			retval=0,
				result;
	int			mem_cnt=1;
	char			total_zone[MAX_HOST_NET_ZONE],
				cur_mem[MAX_HOST];
	struct ZoneData_	*zone_ptr = NULL;


	/* safety */
	if(!groupname || !answer_ptr || !zones)
	{
		(void)vrprint.error(-1, "Interal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(answer_ptr->type != TYPE_GROUP)
	{
		(void)vrprint.error(-1, "Internal Error", "expected a group, but got a %d (in: get_group_info).", answer_ptr->type);
		return(-1);
	}

	/* setup the list (allready done in zone_malloc?) */
	d_list_setup(debuglvl, &answer_ptr->GroupList, NULL);
	answer_ptr->group_member_count = 0;

	/* get the members */
	while((result = zf->ask(debuglvl, zone_backend, groupname, "MEMBER", cur_mem, sizeof(cur_mem), TYPE_GROUP, 1)) == 1)
	{
		answer_ptr->group_member_count++;

		snprintf(total_zone, sizeof(total_zone), "%s.%s.%s", cur_mem, answer_ptr->network_name, answer_ptr->zone_name);

		zone_ptr = search_zonedata(debuglvl, zones, total_zone);
		if(zone_ptr == NULL)
		{
			(void)vrprint.warning("Warning", "The member '%s' of group '%s' was not found in memory.", total_zone, groupname);
			answer_ptr->group_member_count--;
		}
		else
		{
			if(zone_ptr->type == TYPE_GROUP)
			{
				(void)vrprint.warning("Warning", "Only hosts can be groupmembers. Member '%s' of '%s' is a group.", zone_ptr->name, groupname);
				answer_ptr->group_member_count--;
			}
			else
			{

				if(zone_ptr->active == 0)
				{
					if(debuglvl >= LOW)
						fprintf(stdout, "get_group_info: %s not active\n", zone_ptr->name);
				}
					
				if(d_list_append(debuglvl, &answer_ptr->GroupList, zone_ptr) != NULL)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "get_group_info: inserting in list succesful.");
					
					mem_cnt++;
				}
				else
				{
					(void)vrprint.error(-1, "Internal Error", "Unable to insert member '%s' into the list of group '%s' (in: get_group_info).", zone_ptr->name, answer_ptr->name);
					return(-1);
				}
			}
		}
	}
	if(result == -1)
	{
		(void)vrprint.error(-1, "Error", "backend returned error when asking for members of group '%s' (in: %s).", answer_ptr, __FUNCTION__);
		answer_ptr->group_member_count = 0;
		retval = -1;
	}

	if(answer_ptr->group_member_count == 0)
	{
		(void)vrprint.warning("Warning", "group '%s' has no members. Deactivating.", answer_ptr->name, groupname);
		answer_ptr->active = 0;
	}
	
	//fprintf(stdout, "get_group_info: ** end **, return=%d\n", retval);
	return(retval);
}


char *
list_to_portopts(const int debuglvl, d_list *dlist, char *option_name)
{
	d_list_node		*d_node = NULL;
	char			options[MAX_OPTIONS_LENGTH] = "",
				oneport[12] = "",
				*return_ptr = NULL;
	struct portdata		*portrange_ptr = NULL;

	if(option_name != NULL)
	{
		/* option name + starting trema */
		snprintf(options, sizeof(options), "%s=\"", option_name);
	}

	/* loop trough the list and strcat */
	for(d_node = dlist->top; d_node; d_node = d_node->next)
	{
		portrange_ptr = d_node->data;

		/* single port */
		if(portrange_ptr->dst_high == -1)
		{
			snprintf(oneport, sizeof(oneport), "%d,", portrange_ptr->dst_low);

			strcat(options, oneport);
		}
		/* range */
		else
		{
			snprintf(oneport, sizeof(oneport), "%d:%d,", portrange_ptr->dst_low, portrange_ptr->dst_high);
			strlcat(options, oneport, sizeof(options));
		}
	}
	/* overwrite the last comma */
	options[strlen(options)-1] = '\0';

	if(option_name != NULL)
	{
		/* the trailing trema */
		strcat(options, "\"");
	}

	if(!(return_ptr = malloc(strlen(options)+1)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(NULL);
	}

	strlcpy(return_ptr, options, strlen(options)+1);

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "list_to_portopts: '%s'.", options);

	return(return_ptr);
}

int
portopts_to_list(const int debuglvl, const char *opt, d_list *dlist)
{
	int			done=0,
				range=0,
				o=0,
				p=0;
	char			option_string[MAX_OPTIONS_LENGTH];
	struct portdata		*portrange_ptr;

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "portopts_to_list: '%s'", opt);

	/* if the first char is a whitespace, bail out */
	if(opt[0] == ' ')
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "portopts_to_list: first char of 'opt' is a whitespace, so we bail out quietly (and without error).");

		return(0);
	}

	for(o=0, p=0; o <= strlen(opt); o++)
	{
		if(opt[o] != '"')
		{
			if(opt[o] == ':')
				range = 1;

			if(opt[o] == ',')
			{
				option_string[p]='\0';
				done = 1;
				p=0;
			}
			else if(opt[o] == '"' || opt[o] == '\0')
			{
				option_string[p]='\0';
				done = 1;
			}
			else
			{
				option_string[p] = opt[o];
				p++;
			}

			if(done == 1)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "portopts_to_list: now trying to insert: %s", option_string);

				if(!(portrange_ptr = malloc(sizeof(struct portdata))))
				{
					(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
					return(-1);
				}
				portrange_ptr->protocol = -1;
				portrange_ptr->src_low = -1;
				portrange_ptr->src_high = -1;
				portrange_ptr->dst_low = -1;
				portrange_ptr->dst_high = -1;

				if(range == 0)
				{
					portrange_ptr->dst_low = atoi(option_string);
					if(portrange_ptr->dst_low <= 0 || portrange_ptr->dst_low > 65535)
					{
						(void)vrprint.error(-1, "Error", "listenport must be 1-65535.");

						free(portrange_ptr);
						return(-1);
					}
				}
				else
				{
					/* split it! */
					if(split_portrange(option_string, &portrange_ptr->dst_low, &portrange_ptr->dst_high) < 0)
					{
						free(portrange_ptr);
						return(-1);
					}

					/* it is not a range after all */
					if(portrange_ptr->dst_high == 0)
						portrange_ptr->dst_high = -1;

					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "portopts_to_list: listen: %d, %d", portrange_ptr->dst_low, portrange_ptr->dst_high);
				}

				/* append to the list */
				if(d_list_append(debuglvl, dlist, portrange_ptr) < 0)
				{
					(void)vrprint.error(-1, "Internal Error", "appending to list failed (in: %s:%d).", __FUNCTION__, __LINE__);

					free(portrange_ptr);
					return(-1);
				}
				done = 0;
			}
		}
	}

	return(0);
}




/* check_active

    Checks if the supplied zoneinfo is active. It does this by calling ask_backend with
    the question 'ACTIVE'.

  return codes:
  -1  error
   0  not active
   1  active

 */
int
check_active(const int debuglvl, char *name, int type)
{
	int	result = 0;
	char	active[4] = "";

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "check_active: ** start **");

	/* safety */
	if(!name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "query_type: %d, name = %s", type, name);

	if(type >= TYPE_TOO_BIG)
	{
		(void)vrprint.error(-1, "Internal Error", "query_type is out of range (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	/* fw active */
	if(strcasecmp(name, "firewall") == 0 || strncasecmp(name, "firewall", 8) == 0)
	{
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug(__FUNCTION__, "'firewall' is always active.");

		return(1);
	}


	/* service */
	if(type == TYPE_SERVICE || type == TYPE_SERVICEGRP)
	{
		result = sf->ask(debuglvl, serv_backend, name, "ACTIVE", active, sizeof(active), type, 0);
	}
	/* interface */
	else if(type == TYPE_INTERFACE)
	{
		result = af->ask(debuglvl, ifac_backend, name, "ACTIVE", active, sizeof(active), type, 0);
	}
	/* zone, network, host, group */
	else if(type == TYPE_ZONE || type == TYPE_NETWORK || type == TYPE_HOST || type == TYPE_GROUP)
	{
		result = zf->ask(debuglvl, zone_backend, name, "ACTIVE", active, sizeof(active), type, 0);
	}
	else
	{
		(void)vrprint.error(-1, "Internal Error", "query_type '%d' is unsupported (in: %s:%d).", type, __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "'%s' (result: %d).", active, result);

	/* if we have an anwser, check it out */
	if(result == 1)
	{
		if(strncasecmp(active, "yes", 3) == 0)
		{
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug(__FUNCTION__, "'%s' is active.", name);

			return(1);
		}
		else
		{
			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNCTION__, "'%s' is not active.", name);

			return(0);
		}
	}
	else if(result == 0)
	{
		(void)vrprint.warning("Warning", "keyword ACTIVE not found in '%s', deactivating.", name);
		return(0);
	}
	else
	{
		(void)vrprint.error(-1, "Error", "ask_backend returned error (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}
}


/*	get_dynamic_ip

	partly ripped from Net-tools 1.60 (c) Phil Blundell philb@gnu.org and
	Bernd Eckenfels net-tools@lina.inka.de

	Returncodes:
		1: ok
		0: not found
		-1: error
 */
int
get_dynamic_ip(const int debuglvl, char *device, char *answer_ptr, size_t size)
{
	int			numreqs = 30;
	struct ifconf		ifc;
	struct ifreq		*ifr_ptr = NULL,
				ifr_struct;
	int			n;
	int			sockfd = 0;
	char			ipaddress[16] = "";
	struct sockaddr		*sa = NULL;
	struct sockaddr_in	*sin = NULL;


	/* safety */
	if(!device || !answer_ptr || size == 0)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* open a socket for ioctl */
	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if(sockfd == -1)
	{
		(void)vrprint.error(-1, "Error", "couldn't open socket: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/* initialize the buf otherwise realloc will freak out (read segv) */
	ifc.ifc_buf = NULL;
	for (;;)
	{
		ifc.ifc_len = sizeof(struct ifreq) * numreqs;
		/* get some mem */
		if(!(ifc.ifc_buf = realloc(ifc.ifc_buf, ifc.ifc_len)))
		{
			(void)vrprint.error(-1, "Error", "realloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
			(void)close(sockfd);

			return(-1);
		}

		/* get the interfaces from the system */
		if(ioctl(sockfd, SIOCGIFCONF, &ifc) < 0)
		{
			(void)vrprint.error(-1, "Error", "ioctl(SIOCGIFCONF) failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
			free(ifc.ifc_buf);
			(void)close(sockfd);

			return(-1);
		}
		if(ifc.ifc_len == sizeof(struct ifreq) * numreqs)
		{
			/* assume it overflowed and try again */
			numreqs += 10;
			continue;
		}
		break;
	}

	ifr_ptr = ifc.ifc_req;
	for(n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq))
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "ifr_ptr->ifr_name: '%s'.", ifr_ptr->ifr_name);

		if(strcmp(device, ifr_ptr->ifr_name) == 0)
		{
			if(strlcpy(ifr_struct.ifr_name, ifr_ptr->ifr_name, sizeof(ifr_struct.ifr_name)) >= sizeof(ifr_struct.ifr_name))
			{
				(void)vrprint.error(-1, "Error", "buffer overflow (in: %s:%d).", __FUNCTION__, __LINE__);
				(void)close(sockfd);
				free(ifc.ifc_buf);
				return(-1);
			}
			/* we only care about IPv4 */
			ifr_struct.ifr_addr.sa_family = AF_INET;

			/* cast to a socketaddr */
			sa = &ifr_struct.ifr_addr;

			if(ioctl(sockfd, SIOCGIFADDR, &ifr_struct) == 0)
			{
				sin = (struct sockaddr_in *)sa;

				/* get the ipaddress into a string */
				if(inet_ntop(AF_INET, &sin->sin_addr, ipaddress, sizeof(ipaddress)) == NULL)
				{
					(void)vrprint.error(-1, "Error", "getting ipaddress for device '%s' failed: %s (in: %s:%d).", device, strerror(errno), __FUNCTION__, __LINE__);
					(void)close(sockfd);
					free(ifc.ifc_buf);

					return(-1);
				}

				if(debuglvl >= LOW)
					(void)vrprint.debug(__FUNCTION__, ", device: '%s', ipaddress: '%s'.", device, ipaddress);

				/* copy back to the caller */
				if(strlcpy(answer_ptr, ipaddress, size) >= size)
				{
					(void)vrprint.error(-1, "Error", "copying ipaddress for device '%s' failed: destination buffer too small (in: %s:%d).", device, __FUNCTION__, __LINE__);
					(void)close(sockfd);
					free(ifc.ifc_buf);

					return(-1);
				}

				/* found! */
				(void)close(sockfd);
				free(ifc.ifc_buf);
				return(1);
			}
		}
		ifr_ptr++;
	}

	/* not found */
	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "device '%s' not found.", device);

	close(sockfd);
	free(ifc.ifc_buf);
	return(0);
}


/*	check_ipv4address

	Checks if a ipaddress is valid, in two steps
		1: check if the ipaddress itself is valid
		2: check if it belong to the network/netmask combo supplied

	network and netmask may be NULL. If this is the case 'ipaddress' will
	only be checked for validity, not for belonging in a network.
	
	if 'quiet' is 1, we don't print errors, just return the code.

	Returncodes:
		 1: valid
		 0: error: not in network/netmask
		-1: error: no valid ip/net/mask
*/
int
check_ipv4address(const int debuglvl, char *network, char *netmask, char *ipaddress, char quiet)
{
	int			retval = 0;

	struct in_addr		ip;      /* the ipaddress we want to check */
	struct in_addr		net;     /* the network address against we want to check */
	struct in_addr		mask;    /* the netmask of the network */
	struct in_addr		broad;   /* the broadcast address of this network */

	unsigned long int	netmaskvalue = 0;
	unsigned long int	ipaddressvalue = 0;
	unsigned long int	networkvalue = 0;

	unsigned long int	high;
	unsigned long int	low;
	unsigned long int	current;


	/* safety */
	if(!ipaddress)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}
	
	/* first check if the ipaddress itself is valid */
	if(inet_aton(ipaddress, &ip) == 0)
	{
		if(!quiet)
		{
			(void)vrprint.error(-1, "Error", "invalid ipaddress: '%s' (in: %s).", ipaddress, __FUNCTION__);
		}

		return(-1);
	}
	else
	{
		ipaddressvalue = ntohl(ip.s_addr);
		
		if(debuglvl >= HIGH)
			(void)(void)vrprint.debug("Debug", "check_ip: ipaddress = %s", inet_ntoa(ip));

		/* if were only checking ipaddress were happy now. */
		if(!network && !netmask)
			return(1);
	}

	/* check if the networkadress is valid */
	if(inet_aton(network, &net) == 0)
	{
		(void)vrprint.error(-1, "Error", "invalid network: '%s' (in: %s).", network, __FUNCTION__);
		return(-1);
	}
	else
	{
		networkvalue = ntohl(net.s_addr);

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "check_ip: network = %s", inet_ntoa(net));
	}

	/* check if the netmask is valid */
	if(inet_aton(netmask, &mask) == 0)
	{
		(void)vrprint.error(-1, "Error", "invalid netmask: '%s' (in: %s).", netmask, __FUNCTION__);
		return(-1);
	}
	else
	{
		netmaskvalue = ntohl(mask.s_addr);
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "check_ip: netmask = %s", inet_ntoa(mask));
	}


	broad = net;
	broad.s_addr|=~ntohl(netmaskvalue);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "check_ip: broad = %s", inet_ntoa(broad));

	/* get the lowest possible ip in this network/netmask combi */
	low = ntohl(net.s_addr);
	/* get the highest possible ip in this network/netmask combi */
	high = ntohl(broad.s_addr);
	/* get our ip */
	current = ntohl(ip.s_addr);

	/* finally check if the ipaddress fits in the range */
	if(current > low && current < high)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "check_ip: ipaddress %s belongs to network %s with netmask %s", ipaddress, network, netmask);

		retval = 1;
	}

	return(retval);
}


/*	get_mac_address

	Gets the mac address of a host from the backend and checks it.

	Returncodes:
		 0: ok
		-1: error
*/
int
get_mac_address(const int debuglvl, char *hostname, char *answer_ptr, size_t size, regex_t *mac_rgx)
{
	int	retval = 0,
		result = 0;

	/* safety */
	if(!hostname || !mac_rgx)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* ask the backend */
	result = zf->ask(debuglvl, zone_backend, hostname, "MAC", answer_ptr, size, TYPE_HOST, 0);
	if(result == 1)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "get_mac_address: found!");

		retval=1;

		if(strcmp(answer_ptr, "none") == 0)
		{
			retval = 0;
		}
		else
		{
			/* test against the regex */
			if(regexec(mac_rgx, answer_ptr, 0, NULL, 0) != 0)
			{
				(void)vrprint.error(-1, "Error", "MAC '%s' for host '%s' is invalid.", answer_ptr, hostname);
				retval=-1;
			}
		}
	}
	else if(result == 0)
	{
		if(debuglvl >= HIGH) (void)vrprint.debug("Debug", "get_mac_address: not found");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "getting macaddress for %s failed (in: get_mac_address).", hostname);
		retval=-1;
	}

	return(retval);
}


/*
	returncodes
		 0: ok
		-1: error
*/
int
get_danger_info(const int debuglvl, char *danger, char *source, struct danger_info *danger_struct)
{
	/* safety */
	if(!danger || !source || !danger_struct)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "** start **");

	/* spoofing dangers */
	if(strcasecmp(danger, "spoofing") == 0)
	{
		strlcpy(danger_struct->type, "spoof", sizeof(danger_struct->type));

		if(strlcpy(danger_struct->source, source, sizeof(danger_struct->source)) > sizeof(danger_struct->source))
		{
			(void)vrprint.error(-1, "Internal Error", "buffer overflow (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(strcasecmp(source, "loopback") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "127.0.0.0");
			strcpy(danger_struct->source_ip.netmask, "255.0.0.0");
		}
		else if(strcasecmp(source, "class-a") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "10.0.0.0");
			strcpy(danger_struct->source_ip.netmask, "255.0.0.0");
		}
		else if(strcasecmp(source, "class-b") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "172.16.0.0");
			strcpy(danger_struct->source_ip.netmask, "255.240.0.0");
		}
		else if(strcasecmp(source, "class-c") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "192.168.0.0");
			strcpy(danger_struct->source_ip.netmask, "255.255.0.0");
		}
		else if(strcasecmp(source, "class-d") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "224.0.0.0");
			strcpy(danger_struct->source_ip.netmask, "240.0.0.0");
		}
		else if(strcasecmp(source, "class-e") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "240.0.0.0");
			strcpy(danger_struct->source_ip.netmask, "248.0.0.0");
		}
		else if(strcasecmp(source, "test-net") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "192.0.2.0");
			strcpy(danger_struct->source_ip.netmask, "255.255.255.0");
		}
		else if(strcasecmp(source, "lnk-loc-net") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "169.254.0.0");
			strcpy(danger_struct->source_ip.netmask, "255.255.0.0");
		}
		else if(strcasecmp(source, "iana-0/8") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "0.0.0.0");
			strcpy(danger_struct->source_ip.netmask, "255.0.0.0");
		}
		else if(strcasecmp(source, "brdcst-src") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "0.0.0.0");
			strcpy(danger_struct->source_ip.netmask, "255.255.255.255");
		}
		else if(strcasecmp(source, "brdcst-dst") == 0)
		{
			danger_struct->solution = PROT_IPTABLES;
			strcpy(danger_struct->source_ip.ipaddress, "255.255.255.255");
			strcpy(danger_struct->source_ip.netmask, "255.255.255.255");
		}
		else
		{
			(void)vrprint.error(-1, "Internal Error", "unknown source: '%s' (in: %s).", source, __FUNCTION__);
			return(-1);
		}

	}

	/* system dangers */
	else if(strcasecmp(danger, "syn-flood") == 0)
	{
		danger_struct->solution = PROT_PROC_SYS;
		strcpy(danger_struct->proc_entry, "/proc/sys/net/ipv4/tcp_syncookies");
		danger_struct->proc_set_on = 1;
		danger_struct->proc_set_off = 0;
	}
	else if(strcasecmp(danger, "echo-broadcast") == 0)
	{
		danger_struct->solution = PROT_PROC_SYS;
		strcpy(danger_struct->proc_entry, "/proc/sys/net/ipv4/icmp_echo_ignore_broadcasts");
		danger_struct->proc_set_on = 1;
		danger_struct->proc_set_off = 0;
	}

	/* interface dangers */
	else if(strcasecmp(danger, "source-routed-packets") == 0)
	{
		danger_struct->solution = PROT_PROC_INT;
		strcpy(danger_struct->proc_entry, "/proc/sys/net/ipv4/conf/*/accept_source_route");
		danger_struct->proc_set_on = 0;
		danger_struct->proc_set_off = 1;
	}
	else if(strcasecmp(danger, "icmp-redirect") == 0)
	{
		danger_struct->solution = PROT_PROC_INT;
		strcpy(danger_struct->proc_entry, "/proc/sys/net/ipv4/conf/*/accept_redirects");
		danger_struct->proc_set_on = 0;
		danger_struct->proc_set_off = 1;
	}
	else if(strcasecmp(danger, "send-redirect") == 0)
	{
		danger_struct->solution = PROT_PROC_INT;
		strcpy(danger_struct->proc_entry, "/proc/sys/net/ipv4/conf/*/send_redirects");
		danger_struct->proc_set_on = 0;
		danger_struct->proc_set_off = 1;
	}
	else if(strcasecmp(danger, "rp-filter") == 0)
	{
		danger_struct->solution = PROT_PROC_INT;
		strcpy(danger_struct->proc_entry, "/proc/sys/net/ipv4/conf/*/rp_filter");
		danger_struct->proc_set_on = 1;
		danger_struct->proc_set_off = 0;
	}
	else if(strcasecmp(danger, "log-martians") == 0)
	{
		danger_struct->solution = PROT_PROC_INT;
		strcpy(danger_struct->proc_entry, "/proc/sys/net/ipv4/conf/*/log_martians");
		danger_struct->proc_set_on = 1;
		danger_struct->proc_set_off = 0;
	}

	/* default case */
	else
	{
		(void)vrprint.error(-1, "Internal Error", "unknown danger: '%s' (in: %s:%d).", source, __FUNCTION__, __LINE__);
		return(-1);
	}

	return(0);
}


/*	get_network_for_ipv4

	This functions checks to which network a ipv4 address belongs. It does this by
	looping trough the zoneslist and checking if the ipaddress fits in a network
	it returns a pointer to the name of the network, or else NULL.

	CAUTION!!! -> there is one major problem with this aproach. Its can be so
	slooooowwwwww!

	The Solution is providing a list of only networks...
*/
char
*get_network_for_ipv4(const int debuglvl, const char *ipaddress, d_list *zonelist)
{
	struct in_addr		ip;      /* the ipaddress we want to check */
	struct in_addr		net;     /* the network address against we want to check */
	struct in_addr		mask;    /* the netmask of the network */
	struct in_addr		broad;   /* the broadcast address of this network */

	unsigned long int	netmaskvalue = 0;
	unsigned long int	ipaddressvalue = 0;
	unsigned long int	networkvalue = 0;

	unsigned long int	high = 0;
	unsigned long int	low = 0;
	unsigned long int	current = 0;

	unsigned long int	best_so_far = 0;

	struct ZoneData_	*zone_ptr = NULL,
				*best_so_far_ptr = NULL;

	char			*result_ptr = NULL;
	d_list_node		*d_node = NULL;

	/*
		safety
	*/
	if(!ipaddress || !zonelist)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(NULL);
	}


	/* we don't want the local loopback */
	if(strncmp(ipaddress, "127.", 4) == 0)
		return(NULL);

	/* first check if the ipaddress itself is valid */
	if(inet_aton(ipaddress, &ip) == 0)
		return(NULL);

	ipaddressvalue = ntohl(ip.s_addr);

	/* get our ip */
	current = ntohl(ip.s_addr);

	/*
		now loop trough the zonelist in search of networks
	*/
	for(d_node = zonelist->top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(NULL);
		}

		if(zone_ptr->type == TYPE_NETWORK)
		{
			/* check if the networkadress is valid */
			if(inet_aton(zone_ptr->ipv4.network, &net) != 0)
			{
				networkvalue = ntohl(net.s_addr);

				/* check if the netmask is valid */
				if(inet_aton(zone_ptr->ipv4.netmask, &mask) != 0)
				{
					netmaskvalue = ntohl(mask.s_addr);

					/* create broadcast */
					broad = net;
					broad.s_addr|=~ntohl(netmaskvalue);

					/* get the lowest possible ip in this network/netmask combi */
					low = ntohl(net.s_addr);

					/* get the highest possible ip in this network/netmask combi */
					high = ntohl(broad.s_addr);

					/* include the broadcastaddress, so current <= high */
					if((current > low) && (current <= high))
					{
						if(best_so_far == 0 || (high - low) < best_so_far)
						{
							best_so_far = high - low;
							best_so_far_ptr = zone_ptr;
						}
					}
				}
				else
				{
					(void)vrprint.warning("Warning", "network '%s' has an invalid netmask '%s' (in: %s).", zone_ptr->name, zone_ptr->ipv4.netmask, __FUNCTION__);
				}
			}
			else
			{
				(void)vrprint.warning("Warning", "network '%s' has an invalid network address '%s' (in: %s).", zone_ptr->name, zone_ptr->ipv4.network, __FUNCTION__);
			}
		}
	}

	if(best_so_far_ptr != NULL)
	{
		if(!(result_ptr = (char *)malloc(sizeof(best_so_far_ptr->name))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
			return(NULL);
		}

		if(strlcpy(result_ptr, best_so_far_ptr->name, sizeof(best_so_far_ptr->name)) > sizeof(best_so_far_ptr->name))
		{
			(void)vrprint.error(-1, "Internal Error", "buffer overflow! (in: %s).", __FUNCTION__);
			free(result_ptr);

			return(NULL);
		}
	}

	return(result_ptr);
}
