/***************************************************************************
 *   Copyright (C) 2002-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
	functions for detecting the capabilities of Iptables on the
	system.
*/

#include "vuurmuur.h"


static int
iptcap_get_one_cap_from_proc(const int debuglvl, char *procpath, char *request)
{
	char	line[64] = "";
	FILE	*fp = NULL;
	int	retval = 0;

	/* safety */
	if(!procpath || !request)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* open the matches */
	if(!(fp = fopen(procpath, "r")))
	{
		(void)vrprint.error(-1, "Error", "could not open '%s': %s (in: %s:%d).", procpath, strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/* now loop through the file */
	while(fgets(line, sizeof(line), fp) != NULL)
	{
		/* strip the newline if there is one */
		if(line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "%s: '%s'.", procpath, line);

		/* compare the line with the request */
		if(strcmp(line, request) == 0)
		{
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug(__FUNCTION__, "%s: '%s' match!.", procpath, line);

			retval = 1;
			break;
		}
	}

	/* close the file */
	if(fclose(fp) < 0)
	{
		(void)vrprint.error(-1, "Error", "could not close '%s': %s (in: %s:%d).", procpath, strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/* return retval, 1 if found, 0 if not found */
	return(retval);
}


/*
	-1: error
	 0: ok
*/
static int
iptcap_load_module(const int debuglvl, char *modulename)
{
	char	cmd[128] = "";

	/* safety */
	if(!modulename)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* create the commandstring */
	if(snprintf(cmd, sizeof(cmd), "%s -q %s &> /dev/null", conf.modprobe_location, modulename) >= sizeof(cmd))
	{
		(void)vrprint.error(-1, "Error", "commandstring overflow (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}
	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "cmd: '%s'.", cmd);

	/* now execute the command */
	if(pipe_command(debuglvl, cmd, PIPE_QUIET) < 0)
	{
		if(debuglvl >= LOW)
			(void)vrprint.debug(__FUNCTION__, "loading module '%s' failed.", modulename);

		return(-1);
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "loading module '%s' success.", modulename);

	return(0);
}


static int
iptcap_check_cap(const int debuglvl, char *procpath, char *request, char *modulename, char load)
{
	int	result = 0;

	/* safety */
	if(!procpath || !request || !modulename)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* get the cap */
	result = iptcap_get_one_cap_from_proc(debuglvl, procpath, request);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "getting iptcap for '%s' failed (in: %s:%d).", request, __FUNCTION__, __LINE__);
		return(-1);
	}
	else if(result == 0)
	{
		if(debuglvl >= LOW)
			(void)vrprint.debug(__FUNCTION__, "'%s' not loaded or not supported.", request);
	}
	else
	{
		if(debuglvl >= LOW)
			(void)vrprint.debug(__FUNCTION__, "'%s' supported and loaded.", request);

		/* and done :-) */
		return(1);
	}

	/* if load == 0 we bail out now */
	if(load == 0)
		return(0);

	/* try to load the module, if it fails, return 0 */
	result = iptcap_load_module(debuglvl, modulename);
	if(result == -1)
		return(0);

	/* get the cap again */
	result = iptcap_get_one_cap_from_proc(debuglvl, procpath, request);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "getting iptcap for '%s' failed (in: %s:%d).", request, __FUNCTION__, __LINE__);
		return(-1);
	}
	else if(result == 0)
	{
		(void)vrprint.debug(__FUNCTION__, "'%s' not supported.", request);
	}
	else
	{
		(void)vrprint.debug(__FUNCTION__, "'%s' supported and loaded.", request);

		/* and done :-) */
		return(1);
	}

	return(0);
}


/*
	return 1 if file exists, 0 if not.
*/
static int
iptcap_check_file(const int debuglvl, char *path)
{
	FILE	*fp = NULL;

	/* safety */
	if(!path)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(!(fp = fopen(path, "r")))
		return(0);

	fclose(fp);
	return(1);
}


static int
iptcap_get_queue_peer_pid(const int debuglvl, IptCap *iptcap)
{
	char	proc_net_ipqueue[] = "/proc/net/ip_queue",
		line[128] = "",
		pid_number[16] = "";
	FILE	*fp = NULL;
	int	i = 0,
		k = 0,
		copy_now = 0;

	/* safety */
	if(!iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* open the file */
	if(!(fp = fopen(proc_net_ipqueue, "r")))
	{
		(void)vrprint.error(-1, "Error", "opening file '%s' failed: %s (in: %s:%d).", proc_net_ipqueue, strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/* get the first line */
	if(fgets(line, sizeof(line), fp) == NULL)
	{
		(void)vrprint.error(-1, "Error", "reading line from ip_queue failed (in: %s:%d).", __FUNCTION__, __LINE__);

		fclose(fp);
		return(-1);
	}
	/* strip newline */
	if(line[strlen(line)-1] == '\n')
		line[strlen(line)-1] = '\0';

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "line '%s'.", line);

	/* interpret the line */
	for(i = 0, k = 0; i < strlen(line); i++)
	{
		if(copy_now)
		{
			pid_number[k] = line[i];

			k++;
		}

		if(line[i] == ':')
			copy_now = 1;
	}
	/* term */
	pid_number[k] = '\0';

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "pid_number '%s'.", pid_number);

	/* convert */
	iptcap->queue_peer_pid = atoi(pid_number);

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "pid '%u'.", iptcap->queue_peer_pid);

	return(0);
}


int
check_iptcaps(const int debuglvl, IptCap *iptcap, char load_modules)
{
	int	result = 0;

	/* safety */
	if(!iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* load the caps */
	result = load_iptcaps(debuglvl, iptcap, load_modules);
	if(result == -1)
	{
		(void)vrprint.error(-1, "Error", "loading iptables capabilities failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(!iptcap->proc_net_names)
	{
		(void)vrprint.warning("Warning", "'/proc/net/ip_tables_names' missing: no iptables-support in the kernel?");

		/* assume yes */
		iptcap->table_filter = 1;
		iptcap->table_mangle = 1;
		iptcap->table_nat    = 1;
	}
	if(!iptcap->proc_net_targets)
	{
		(void)vrprint.warning("Warning", "'/proc/net/ip_tables_targets' missing: no iptables-support in the kernel?");

		/* assume yes */
		iptcap->target_snat = 1;
		iptcap->target_dnat = 1;

		iptcap->target_reject = 1;
		iptcap->target_log = 1;
		iptcap->target_redirect = 1;
		iptcap->target_mark = 1;
		iptcap->target_masquerade = 1;
	}
	if(!iptcap->proc_net_matches)
	{
		(void)vrprint.warning("Warning", "'/proc/net/ip_tables_matches' missing: no iptables-support in the kernel?");

		/* assume yes */
		iptcap->match_tcp = 1;
		iptcap->match_udp = 1;
		iptcap->match_icmp = 1;

		iptcap->match_mark = 1;
		iptcap->match_state = 1;
		iptcap->match_helper = 1;
		iptcap->match_length = 1;
		iptcap->match_limit = 1;
		iptcap->match_mac = 1;
	}

	if(iptcap->proc_net_names && (!iptcap->table_filter || !iptcap->table_mangle || !iptcap->table_nat))
	{
		(void)vrprint.error(-1, "Error", "no iptables-support in the kernel (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	if(!iptcap->conntrack)
	{
		(void)vrprint.error(-1, "Error", "no connection tracking support in the kernel (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	if(iptcap->proc_net_matches && (!iptcap->match_tcp || !iptcap->match_udp || !iptcap->match_icmp))
	{
		(void)vrprint.error(-1, "Error", "incomplete iptables-support in the kernel: tcp, udp or icmp support missing (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(iptcap->proc_net_matches && !iptcap->match_state)
	{
		(void)vrprint.error(-1, "Error", "incomplete iptables-support in the kernel: state support missing (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(iptcap->target_queue && iptcap->queue_peer_pid == 0)
	{
		(void)vrprint.warning("Warning", "no application is currently listening to the queue. Queue rules may be uneffective.");
	}

	return(0);
}


int
load_iptcaps(const int debuglvl, IptCap *iptcap, char load_modules)
{
	char	proc_net_match[]	= "/proc/net/ip_tables_matches",
		proc_net_target[]	= "/proc/net/ip_tables_targets",
		proc_net_names[]	= "/proc/net/ip_tables_names",
		proc_net_ipqueue[]	= "/proc/net/ip_queue",
		proc_net_ipconntrack[]	= "/proc/net/ip_conntrack";
	int	result = 0;


	/* safety */
	if(!iptcap)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}
	/* init */
	memset(iptcap, 0, sizeof(IptCap));

	/*
		PROC FILES
	*/
	if(!(iptcap_check_file(debuglvl, proc_net_match)))
	{
		if(load_modules == 1)
		{
			/* try to load the module, if it fails, return 0 */
			result = iptcap_load_module(debuglvl, "ip_tables");
			if(result == 1)	iptcap->proc_net_matches = 1;
			else		iptcap->proc_net_matches = 0;
		}
	}
	else
	{
		iptcap->proc_net_matches = 1;
	}
	if(!(iptcap_check_file(debuglvl, proc_net_target)))
	{
		if(load_modules == 1)
		{
			/* try to load the module, if it fails, return 0 */
			result = iptcap_load_module(debuglvl, "ip_tables");
			if(result == 1)	iptcap->proc_net_targets = 1;
			else		iptcap->proc_net_targets = 0;
		}
	}
	else
	{
		iptcap->proc_net_targets = 1;
	}
	if(!(iptcap_check_file(debuglvl, proc_net_names)))
	{
		if(load_modules == 1)
		{
			/* try to load the module, if it fails, return 0 */
			result = iptcap_load_module(debuglvl, "ip_tables");
			if(result == 1)	iptcap->proc_net_names = 1;
			else		iptcap->proc_net_names = 0;
		}
	}
	else
	{
		iptcap->proc_net_names = 1;
	}

	/*
		NAMES
	*/
	if(!iptcap->proc_net_names)
		return(0);

	result = iptcap_check_cap(debuglvl, proc_net_names, "filter", "iptable_filter", load_modules);
	if(result == 1)	iptcap->table_filter = 1;
	else		iptcap->table_filter = 0;

	result = iptcap_check_cap(debuglvl, proc_net_names, "mangle", "iptable_mangle", load_modules);
	if(result == 1)	iptcap->table_mangle = 1;
	else		iptcap->table_mangle = 0;

	result = iptcap_check_cap(debuglvl, proc_net_names, "nat", "iptable_nat", load_modules);
	if(result == 1)	iptcap->table_nat = 1;
	else		iptcap->table_nat = 0;


	/* check for the CONNTRACK */
	if(!(iptcap_check_file(debuglvl, proc_net_ipconntrack)))
	{
		if(load_modules == 1)
		{
			/* try to load the module, if it fails, return 0 */
			(void)iptcap_load_module(debuglvl, "ip_conntrack");

			/* check again */
			if(!(iptcap_check_file(debuglvl, proc_net_ipconntrack)))
				iptcap->conntrack = 0;
			else
				iptcap->conntrack = 1;
		}
	}
	else
	{
		iptcap->conntrack = 1;
	}

	/*
		MATCHES (uncapitalized)
	*/
	if(!iptcap->proc_net_matches)
		return(0);

	/* tcp */
	result = iptcap_check_cap(debuglvl, proc_net_match, "tcp", "ip_tables", load_modules);
	if(result == 1)	iptcap->match_tcp = 1;
	else		iptcap->match_tcp = 0;

	/* udp */
	result = iptcap_check_cap(debuglvl, proc_net_match, "udp", "ip_tables", load_modules);
	if(result == 1)	iptcap->match_udp = 1;
	else		iptcap->match_udp = 0;

	/* icmp */
	result = iptcap_check_cap(debuglvl, proc_net_match, "icmp", "ip_tables", load_modules);
	if(result == 1)	iptcap->match_icmp = 1;
	else		iptcap->match_icmp = 0;

	/* state match */
	result = iptcap_check_cap(debuglvl, proc_net_match, "state", "ipt_state", load_modules);
	if(result == 1)	iptcap->match_state = 1;
	else		iptcap->match_state = 0;


	/* length match */
	result = iptcap_check_cap(debuglvl, proc_net_match, "length", "ipt_length", load_modules);
	if(result == 1)	iptcap->match_length = 1;
	else		iptcap->match_length = 0;

	/* limit match */
	result = iptcap_check_cap(debuglvl, proc_net_match, "limit", "ipt_limit", load_modules);
	if(result == 1)	iptcap->match_limit = 1;
	else		iptcap->match_limit = 0;

	/* mark match */
	result = iptcap_check_cap(debuglvl, proc_net_match, "mark", "ipt_mark", load_modules);
	if(result == 1)	iptcap->match_mark = 1;
	else		iptcap->match_mark = 0;

	/* mac match */
	result = iptcap_check_cap(debuglvl, proc_net_match, "mac", "ipt_mac", load_modules);
	if(result == 1)	iptcap->match_mac = 1;
	else		iptcap->match_mac = 0;

	/* helper match */
	result = iptcap_check_cap(debuglvl, proc_net_match, "helper", "ipt_helper", load_modules);
	if(result == 1)	iptcap->match_helper = 1;
	else		iptcap->match_helper = 0;

	/*
		TARGETS (capitalized)
	*/
	if(!iptcap->proc_net_targets)
		return(0);

	/* DNAT target */
	result = iptcap_check_cap(debuglvl, proc_net_target, "DNAT", "iptable_nat", load_modules);
	if(result == 1)	iptcap->target_dnat = 1;
	else		iptcap->target_dnat = 0;

	/* SNAT target */
	result = iptcap_check_cap(debuglvl, proc_net_target, "SNAT", "iptable_nat", load_modules);
	if(result == 1)	iptcap->target_snat = 1;
	else		iptcap->target_snat = 0;


	/* REJECT target */
	result = iptcap_check_cap(debuglvl, proc_net_target, "REJECT", "ipt_REJECT", load_modules);
	if(result == 1)	iptcap->target_reject = 1;
	else		iptcap->target_reject = 0;

	/* LOG target */
	result = iptcap_check_cap(debuglvl, proc_net_target, "LOG", "ipt_LOG", load_modules);
	if(result == 1)	iptcap->target_log = 1;
	else		iptcap->target_log = 0;

	/* REDIRECT target */
	result = iptcap_check_cap(debuglvl, proc_net_target, "REDIRECT", "ipt_REDIRECT", load_modules);
	if(result == 1)	iptcap->target_redirect = 1;
	else		iptcap->target_redirect = 0;

	/* MARK target */
	result = iptcap_check_cap(debuglvl, proc_net_target, "MARK", "ipt_MARK", load_modules);
	if(result == 1)	iptcap->target_mark = 1;
	else		iptcap->target_mark = 0;

	/* MASQUERADE target */
	result = iptcap_check_cap(debuglvl, proc_net_target, "MASQUERADE", "ipt_MASQUERADE", load_modules);
	if(result == 1)	iptcap->target_masquerade = 1;
	else		iptcap->target_masquerade = 0;


	/* finally check for the QUEUE */
	if(!(iptcap_check_file(debuglvl, proc_net_ipqueue)))
	{
		if(load_modules == 1)
		{
			/* try to load the module, if it fails, return 0 */
			(void)iptcap_load_module(debuglvl, "ip_queue");

			/* check again */
			if(!(iptcap_check_file(debuglvl, proc_net_ipqueue)))
				iptcap->target_queue = 0;
			else
			{
				iptcap->target_queue = 1;
				
				if(iptcap_get_queue_peer_pid(debuglvl, iptcap) < 0)
				{
					(void)vrprint.error(-1, "Error", "getting queue peer pid failed (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
			}

		}
	}
	else
	{
		iptcap->target_queue = 1;

		if(iptcap_get_queue_peer_pid(debuglvl, iptcap) < 0)
		{
			(void)vrprint.error(-1, "Error", "getting queue peer pid failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	return(0);
}
