/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
/***************************************************************************
 *   In this file most of the zonedata manipulation functions reside       *
 ***************************************************************************/


#include "vuurmuur.h"

/*	insert_zonedata_list

	Inserts a zone into the list. It sorts by name. It makes sure the datastructure is
	preserved:
		zone dd
			network ee
				host	aa
				host	bb
				host	zz
				group	cc
				group	xx
			network gg
		zone ee
			etc.

	Returncodes:
		 0: ok
		-1: (serious) error
*/
int
insert_zonedata_list(const int debuglvl, Zones *zones, const struct ZoneData_ *zone_ptr)
{
	struct ZoneData_	*check_zone_ptr = NULL,
				*cur_zone = NULL,
				*cur_network = NULL;
	int			insert_here = 0,
				in_the_right_scope = 0;
	d_list_node		*d_node = NULL;


	/* safety first */
	if(!zones || !zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* if the list is empty, just insert */
	if(zones->list.len == 0)
		insert_here = 1;
	else
	{
		/* loop trough the existing list to see where to insert */
		for(d_node = zones->list.top; d_node && !insert_here; d_node = d_node->next)
		{
			if(!(check_zone_ptr = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: check_zone_ptr: name: %s, type: %d.", __FUNCTION__, check_zone_ptr->name, check_zone_ptr->type);

			/* store the last zone and network so we can determine the scope */
			if(check_zone_ptr->type == TYPE_ZONE)
				cur_zone = check_zone_ptr;
			else if(check_zone_ptr->type == TYPE_NETWORK)
				cur_network = check_zone_ptr;

			/* see if we are in the right scope: this means that a host is in its 'own' network, a network in its 'own' zone, etc. */
			if(	(zone_ptr->type == TYPE_ZONE) ||
				(zone_ptr->type == TYPE_NETWORK && cur_zone && strcmp(cur_zone->zone_name, zone_ptr->zone_name) == 0) ||
				(zone_ptr->type == TYPE_HOST    && cur_zone && strcmp(cur_zone->zone_name, zone_ptr->zone_name) == 0  && cur_network && strcmp(cur_network->network_name, zone_ptr->network_name) == 0) ||
				(zone_ptr->type == TYPE_GROUP   && cur_zone && strcmp(cur_zone->zone_name, zone_ptr->zone_name) == 0  && cur_network && strcmp(cur_network->network_name, zone_ptr->network_name) == 0)
			)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: in the right scope %s", __FUNCTION__, zone_ptr->name);

				/* we are in the right scope */
				in_the_right_scope = 1;

				/* only compare with our own type (racists! ;) */
				if(zone_ptr->type == check_zone_ptr->type)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "%s: same type %s", __FUNCTION__, zone_ptr->name);

					/*	now compare the name.

						if the name is the same (should never happen) or 'smaller', insert
						before the current item
					*/
					if(strcmp(zone_ptr->name, check_zone_ptr->name) <= 0)
					{
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "%s: insert here %s", __FUNCTION__, zone_ptr->name);

						insert_here = 1;
						break;
					}
				}
			}
			/*	if were not in the right scope anymore, we need to insert right now!
				This makes sure the data is inserted in our scope. We get here when
				for example inserting the first host in a network. It can never match the
				type comparison above.
			*/
			else if(in_the_right_scope)
			{
				insert_here = 1;
				break;
			}
		}
	}

	/* is d_node is untouched (NULL) we prepend. */
	if(insert_here && !d_node)
	{
		/* prepend */
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: prepend %s", __FUNCTION__, zone_ptr->name);

		if(d_list_prepend(debuglvl, &zones->list, zone_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "prepending zonedata into the list failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}
	else if(insert_here && d_node)
	{
		/* insert before */
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: insert %s", __FUNCTION__, zone_ptr->name);

		if(d_list_insert_before(debuglvl, &zones->list, d_node, zone_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "inserting zonedata into the list failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}
	else
	{
		/* append */
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: append %s", __FUNCTION__, zone_ptr->name);

		if(d_list_append(debuglvl, &zones->list, zone_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending zonedata into the list failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	/* for debugging, print the entire list to the log */
	if(debuglvl >= HIGH)
	{
		for(d_node = zones->list.top; d_node; d_node = d_node->next)
		{
			if(!(check_zone_ptr = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
				return(-1);
			}

			(void)vrprint.debug("Debug", "%s: list: check_zone_ptr: name: %s, type: %d.", __FUNCTION__, check_zone_ptr->name, check_zone_ptr->type);
		}
	}

	return(0);
}

/*	insert_zonedata

	Inserts the zonedata 'name' into the linked-list.

	Returncodes:
		-1: error
		 0: succes
*/
int
insert_zonedata(const int debuglvl, Zones *zones, Interfaces *interfaces, char *name, int type, struct rgx_ *reg)
{
	struct ZoneData_	*zone_ptr = NULL;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "insert_zonedata: ** start **");

	/*
		please put on your safetybelt
	*/
	if(!zones || !name || !reg || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/*
		claiming the memory we need, in case of error zone_malloc will tell the user
	*/
	if(!(zone_ptr = zone_malloc(debuglvl)))
		return(-1);

	/*
		read the data for this zone
	*/
	if(read_zonedata(debuglvl, zones, interfaces, name, type, zone_ptr, reg) < 0)
	{
		(void)vrprint.error(-1, "Error", "Reading the data for zone '%s' failed (in: insert_zonedata).", name);

		free(zone_ptr);
		return(-1);
	}

	/*
		now insert into the list
	*/
	if(insert_zonedata_list(debuglvl, zones, zone_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "inserting zone '%s' failed (in: %s:%d).", name, __FUNCTION__, __LINE__);
		return(-1);
	}

	zone_ptr->status = ST_ADDED;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "insert_zonedata: ** end **");

	return(0);
}

/*	read_zonedata

	Reads all the info for a zone.

	Returncodes:
		 0: ok
		-1: error
*/
int
read_zonedata(const int debuglvl, Zones *zones, Interfaces *interfaces, char *name, int type, struct ZoneData_ *zone_ptr, struct rgx_ *reg)
{
	int	result=0,
		retval=0;
	char	network[MAX_NET_ZONE];


	/* safety */
	if(!name || !zone_ptr || !zones || !reg || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: read_zonedata)");
		return(-1);
	}

	if(	type != TYPE_ZONE &&
		type != TYPE_NETWORK &&
		type != TYPE_HOST &&
		type != TYPE_GROUP)
	{
		(void)vrprint.error(-1, "Interal Error", "wrong zonetype %d (in: %s:%d).", type, __FUNCTION__, __LINE__);
		return(-1);
	}

	if(validate_zonename(debuglvl, name, 1, NULL, NULL, NULL, reg->zonename) != 0)
	{
		(void)vrprint.error(-1, "Error", "invalid zonename: '%s' (in: read_zonedata).", name);
		return(-1);
	}

	/* copy the name to the structure */
	if(strlcpy(zone_ptr->name, name, sizeof(zone_ptr->name)) > sizeof(zone_ptr->name))
	{
		(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: read_zonedata).");
		return(-1);
	}

	/* set the type */
	zone_ptr->type = type;

	/* split and check */
	if(retval == 0)
	{
		result = split_and_check(debuglvl, zones, name, zone_ptr, reg->zonename);
		if(result == 0)
		{
			/* ok */
			zone_ptr->active = 1;
		}
		else if(result == 1)
		{
			/* not active */
			(void)vrprint.info("Info", "Note: Zone '%s' is not active.", name);
			zone_ptr->active = 0;
		}
		else
		{
			/* error */
			(void)vrprint.error(-1, "Error", "splitting and checking zone '%s' failed (in: %s:%d).", name, __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(zone_ptr->type != TYPE_ZONE && zone_ptr->type != TYPE_GROUP)
	{
		// for get_interface
		if(zone_ptr->type == TYPE_HOST)
		{
			snprintf(network, sizeof(network), "%s.%s", zone_ptr->network_name, zone_ptr->zone_name);
		}
		else
		{
			if(strlcpy(network, name, sizeof(network)) > sizeof(network))
			{
				(void)vrprint.error(-1, "Error", "buffer overflow (in: read_zonedata).");

				free(zone_ptr);
				return(-1);
			}
		}

		if(zone_ptr->type == TYPE_NETWORK)
		{
			result = zones_network_get_interfaces(debuglvl, zone_ptr, interfaces);
			if(result < 0)
			{
				(void)vrprint.error(-1, "Error", "getting interfaces for network '%s' failed. Setting ACTIVE to NO for this network.", name);

				zone_ptr->active = 0;
			}

			result = zones_network_get_protectrules(debuglvl, zone_ptr);
			if(result < 0)
			{
				(void)vrprint.error(-1, "Error", "getting protectrules for network '%s' failed. Setting ACTIVE to NO for this network.", name);

				zone_ptr->active = 0;
			}
		}

		/*
			get ip and mask
		*/
		if(retval >= 0)
		{
			result = get_ip_info(debuglvl, name, zone_ptr, reg);
			if(result != 0)
			{
				// error
				(void)vrprint.error(-1, "Error", "get_ip_info for zonedata '%s' failed. Setting ACTIVE to NO for this zone.", name);

				zone_ptr->active = 0;
			}
		}
	}
	else if(zone_ptr->type == TYPE_GROUP)
	{
		// get group info
		if(retval >= 0)
		{
			result = get_group_info(debuglvl, zones, name, zone_ptr);
			if(result != 0)
			{
				// error
				(void)vrprint.error(-1, "Error", "Get_group_info for zonedata '%s' failed.", name);

				zone_ptr->active = 0;
			}
		}
	}

	return(retval);
}

/*	search_zonedata

	Function to search the ServicesList.

	It returns the pointer or a NULL-pointer if not found.
*/
void *
search_zonedata(const int debuglvl, const Zones *zones, char *name)
{
	d_list_node		*d_node = NULL;
	struct ZoneData_	*zonedata_ptr = NULL;


	/* safety */
	if(!name || !zones)
	{
		vrprint.error(-1, "Internal Error", "parameter problem (in: search_zonedata).");
		return(NULL);
	}


	/* now search */
	for(d_node = zones->list.top; d_node ; d_node = d_node->next)
	{
		if(!(zonedata_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(NULL);
		}

		if(strcmp(zonedata_ptr->name, name) == 0)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "search_zonedata: Zone '%s' found.", name);

			/* found, return */
			return(zonedata_ptr);
		}
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug", "search_zonedata: zone '%s' not found in memory.", name);

	/* return NULL pointer */
	return(NULL);
}

/*	search_zonedata

	Function to search the ServicesList.

	It returns the pointer or a NULL-pointer if not found.

void *
search_zonedata_hash(const int debuglvl, const char *name)
{
	//char			*data;
	struct ZoneData_	*search_ptr, *old_search_ptr;
	int			retval;

	if(!name)
		return(NULL);

	search_ptr = malloc(sizeof(struct ZoneData_));
	strcpy(search_ptr->name, name);

	old_search_ptr = search_ptr;

        retval = hash_search(debuglvl,&host_hash_table, (void **)&search_ptr);
        if(retval==-1)
	{
		search_ptr = NULL;
        }

	free(old_search_ptr);

	// return the pointer
	return(search_ptr);
}
*/
/*- print_list - */
void
zonedata_print_list(const Zones *zones)
{
	d_list_node		*d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;

	// print size
	fprintf(stdout, "ZonedataList size: %d\n", zones->list.len);

	for(d_node = zones->list.top; d_node ; d_node = d_node->next)
	{
		zone_ptr = d_node->data;

		fprintf(stdout, "zone: %s, status: %d, active: %d, type: %d\n", zone_ptr->name, zone_ptr->status, zone_ptr->active, zone_ptr->type);
	}

	return;
}

/*	init_zonedata

	Loads all zonedata in memory.

	returncodes:
		 0: succes
		 1: succes with one or more zonedata entries failed
		-1: error

	TODO:
		input validation
*/
int
init_zonedata(const int debuglvl, Zones *zones, Interfaces *interfaces, struct rgx_ *reg)
{
	int	retval=0,
		result,
		counter=0,
		failedcounter=0,
		zonetype=0,
		prev_zonetype=0;
	char	zonename[MAX_HOST_NET_ZONE] = "";

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "** start **");

	
	/* safety */
	if(!zones || !interfaces || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* init */
	memset(zones, 0, sizeof(Zones));

	/* create the list */
	if(d_list_setup(debuglvl, &zones->list, NULL) < 0)
		return(-1);


	/* get the info from the backend */
	while(zf->list(debuglvl, zone_backend, zonename, &zonetype, CAT_ZONES) != NULL)
	{
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug(__FUNCTION__, "loading zone: '%s', type: %d", zonename, zonetype);

		if(validate_zonename(debuglvl, zonename, 1, NULL, NULL, NULL, reg->zonename) == 0)
		{
			result = insert_zonedata(debuglvl, zones, interfaces, zonename, zonetype, reg);
			if(result < 0)
			{
				(void)vrprint.error(-1, "Error", "insert_zonedata failed");
				failedcounter++;

				retval = 1;
			}
			else
			{
				counter++;

				if(debuglvl >= LOW)
					(void)vrprint.debug(__FUNCTION__, "loading zone succes: '%s' (type %d).", zonename, zonetype);
			}
		}
		else
		{
			failedcounter++;
		}
		prev_zonetype = zonetype;
	}

	if(retval != -1)
	{
		(void)vrprint.info("Info", "%d zones loaded, %d failed.", counter, failedcounter);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "** end **, retval=%d", retval);

	return(retval);
}


void
destroy_zonedatalist(const int debuglvl, Zones *zones)
{
	d_list_node		*d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;

	if(!zones)
		return;

	for(d_node = zones->list.top; d_node ; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return;
		}

		zone_free(debuglvl, zone_ptr);
	}

	d_list_cleanup(debuglvl, &zones->list);
}

int
delete_zone(const int debuglvl, Zones *zones, char *zonename, int zonetype)
{
	struct ZoneData_	*zone_ptr = NULL,
				*zone_list_ptr = NULL;
	d_list_node		*d_node = NULL;
	char			name[MAX_HOST_NET_ZONE];

	/* safety */
	if(!zonename || !zones)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: delete_zone).");
		return(-1);
	}

	/* copy the name to an array so we can display the name after the deletion is complete */
	strlcpy(name, zonename, sizeof(name));

	/* check zonetype */
	if(zonetype != TYPE_ZONE && zonetype != TYPE_NETWORK && zonetype != TYPE_HOST && zonetype != TYPE_GROUP)
	{
		(void)vrprint.error(-1, "Internal Error", "expected a zone, network, host or group, but got a %d (in: delete_zone).", zonetype);
		return(-1);
	}

	/* search the zone */
	if(!(zone_ptr = search_zonedata(debuglvl, zones, zonename)))
	{
		(void)vrprint.error(-1, "Internal Error", "zone '%s' not found in memory (in: delete_zone).", zonename);
		return(-1);
	}

	/* delete the zone from the backend */
	if(zf->del(debuglvl, zone_backend, zonename, zonetype, 1) < 0)
	{
		(void)vrprint.error(-1, "Error", "zone '%s' could not be deleted.", zonename);
		return(-1);
	}

	/* find its position in the list */
	for(d_node = zones->list.top; d_node; d_node = d_node->next)
	{
		if(!(zone_list_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "zone_list_ptr == NULL (in: delete_zone).");
			return(-1);
		}

		if(strcmp(zonename, zone_list_ptr->name) == 0)
		{
			/* remove from list */
			if(d_list_remove_node(debuglvl, &zones->list, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "failed to delete the zone from the list (in: delete_zone).");
				return(-1);
			}

			/* remove from memory */
			zone_free(debuglvl, zone_list_ptr);
			
			/* we're done */
			(void)vrprint.info("Info", "Zone '%s' succesfully deleted from the backend.", name);
			return(0);
		}
	}

	/* we should never get here */
	(void)vrprint.error(-1, "Internal Error", "Zone not found in memory.");
	return(-1);
}

/*	new_zone

	TODO: the spliting of name is totally wacked
*/
int
new_zone(const int debuglvl, Zones *zones, char *zonename, int zonetype)
{
	struct ZoneData_	*zone_ptr=NULL;
	int			dotcount=0,
				i=0,
				x=0;
	char			parent_str[MAX_NET_ZONE] = "";

	
	/* safety */
	if(!zonename || !zones)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* allocated memory for the new zone */
	if(!(zone_ptr = zone_malloc(debuglvl)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed (in: new_zone).");
		return(-1);
	}

	for(i=0, dotcount=0; i < strlen(zonename); i++)
	{
		if(zonename[i] == '.')
			dotcount++;
	}

	if(dotcount > 2)
	{
		(void)vrprint.error(-1, "Error", "Invalid name '%s' (in: new_zone).", zonename);
		return(-1);
	}

	if(dotcount == 0)
	{
		for(i=0; i < strlen(zonename); i++)
		{
			zone_ptr->zone_name[i] = zonename[i];
		}
		zone_ptr->zone_name[i]='\0';
	}
	else if(dotcount == 1)
	{
		// network
		for(i=0; i < strlen(zonename);i++)
		{
			if(zonename[i] == '.')
				break;
			zone_ptr->network_name[i] = zonename[i];
		}
		zone_ptr->network_name[i]='\0';

		for(i=strlen(zone_ptr->network_name)+1, x=0; i < strlen(zonename) ;i++, x++)
		{
			zone_ptr->zone_name[x] = zonename[i];
		}
		zone_ptr->zone_name[x]='\0';

	}
	else
	{
		// host/group
		for(i=0; i < strlen(zonename);i++)
		{
			if(zonename[i] == '.')
				break;
			zone_ptr->host_name[i] = zonename[i];
		}
		zone_ptr->host_name[i]='\0';

		for(i=strlen(zone_ptr->host_name)+1, x=0; i < strlen(zonename) ;i++, x++)
		{
			if(zonename[i] == '.')
				break;
			zone_ptr->network_name[x] = zonename[i];
		}
		zone_ptr->network_name[x]='\0';

		for(i = strlen(zone_ptr->host_name) + 1 + strlen(zone_ptr->network_name) + 1, x = 0; i < strlen(zonename); i++, x++)
		{
			zone_ptr->zone_name[x] = zonename[i];
		}
		zone_ptr->zone_name[x]='\0';
	}


	/* check if the zone already exists */
	if(search_zonedata(debuglvl, zones, zonename) != NULL)
	{
		(void)vrprint.error(-1, "Error", "zonename '%s' already exists (in: new_zone).", zonename);

		zone_free(debuglvl, zone_ptr);
		return(-1);
	}


	/* set the bare minimum */
	strlcpy(zone_ptr->name, zonename, sizeof(zone_ptr->name));
	zone_ptr->type = zonetype;


	/* set the parent(s) */
	snprintf(parent_str, sizeof(parent_str), "%s.%s", zone_ptr->network_name, zone_ptr->zone_name);
	if(zone_ptr->type == TYPE_HOST || zone_ptr->type == TYPE_GROUP)
	{
		if(!(zone_ptr->network_parent = search_zonedata(debuglvl, zones, parent_str)))
		{
			(void)vrprint.error(-1, "Internal Error", "can't find the network-parent in the list (in: new_zone).");
			return(-1);
		}
	}
	if(zone_ptr->type == TYPE_HOST || zone_ptr->type == TYPE_GROUP || zone_ptr->type == TYPE_NETWORK)
	{
		if(!(zone_ptr->zone_parent = search_zonedata(debuglvl, zones, zone_ptr->zone_name)))
		{
			(void)vrprint.error(-1, "Internal Error", "can't find the zone-parent in the list (in: new_zone).");
			return(-1);
		}
	}


	/* insert into the list */
	if(insert_zonedata_list(debuglvl, zones, zone_ptr) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "unable to insert new zone into the list (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* add the zone to the backend */
	if(zf->add(debuglvl, zone_backend, zonename, zonetype) < 0)
	{
		(void)vrprint.error(-1, "Error", "Add to backend failed (in: new_zone).");
		return(-1);
	}


	/* set active */
	if(zf->tell(debuglvl, zone_backend, zonename, "ACTIVE", zone_ptr->active ? "Yes" : "No", 1, zonetype) < 0)
	{
		(void)vrprint.error(-1, "Error", "Tell backend failed (in: new_zone).");
		return(-1);
	}

	(void)vrprint.info("Info", "new zone '%s' succesfully added to the backend.", zonename);
	return(0);
}

/*
	TODO: input check
*/
int count_zones(int debuglvl, Zones *zones, int type, char *filter_network, char *filter_zone)
{
	struct ZoneData_	*zone_ptr = NULL;
	int			count = 0;
	d_list_node		*d_node = NULL;

	for(d_node = zones->list.top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			//TODO
			return(-1);
		}

		if(zone_ptr->type == type)
		{
			if(filter_zone != NULL)
			{
				if(strcmp(filter_zone, zone_ptr->zone_name) == 0)
				{
					if(filter_network != NULL)
					{
						if(strcmp(filter_network, zone_ptr->network_name) == 0)
						{
							count++;
						}
					}
					else
					{
						count++;
					}
				}
			}
			else
			{
				count++;
			}
		}
	}

	return(count);
}

/*	zonelist_to_networklist

	Function to load the networks of a list into a new networklist. The
	networks in the original list will be untouched.
	
	Returncodes:
		 0: ok
		-1: error

*/
int
zonelist_to_networklist(const int debuglvl, Zones *zones, d_list *network_list)
{
	d_list_node		*d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;

	/*
		safety
	*/
	if(!zones || !network_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: zonelist_to_networklist).");
		return(-1);
	}

	for(d_node = zones->list.top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "zone_ptr == NULL! (in: zonelist_to_networklist).");
			return(-1);
		}

		if(zone_ptr->type == TYPE_NETWORK)
		{
			if(d_list_append(debuglvl, network_list, zone_ptr) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "appending to the list failed (in: zonelist_to_networklist).");
				return(-1);
			}
		}
	}

	return(0);
}

/*	add_broadcasts_zonelist

	Adds the broadcast address of networks as TYPE_FIREWALL's to the zone_list

	We ignore 255.255.255.255 because its a general broadcast, and i don't want
	it to show like internet.ext(broadcast).

	Returncodes:
		 0: ok
		-1: error
*/
int
add_broadcasts_zonelist(const int debuglvl, Zones *zones)
{
	struct ZoneData_	*zone_ptr = NULL,
				*broadcast_ptr = NULL;
	d_list_node		*d_node = NULL;

	/* safety */
	if(!zones)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/*
		now run through the list
	*/
	for(d_node = zones->list.top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(zone_ptr->type == TYPE_NETWORK)
		{
			if(strcmp(zone_ptr->ipv4.broadcast, "255.255.255.255") != 0)
			{
				/* allocate memory */
				if(!(broadcast_ptr = zone_malloc(debuglvl)))
					return(-1);

				/*	store the bare minimum:
						name
						ipaddress
						type
				*/
				snprintf(broadcast_ptr->name, MAX_HOST_NET_ZONE, "%s(broadcast)", zone_ptr->name);

				strlcpy(broadcast_ptr->ipv4.ipaddress, zone_ptr->ipv4.broadcast, sizeof(broadcast_ptr->ipv4.ipaddress));

				broadcast_ptr->type = TYPE_FIREWALL;

				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "add_broadcasts_zonelist: %s addr: %s", broadcast_ptr->name, broadcast_ptr->ipv4.ipaddress);

				/* insert into the list */
				if(d_list_append(debuglvl, &zones->list, broadcast_ptr) < 0)
				{
					(void)vrprint.error(-1, "Internal Error", "appending to the list failed (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
			}
		}
	}

	return(0);
}

/*
	NOTE: THIS FUCNTION REQUIRES THE ZONE, NETWORK AND HOST VARIABLES TO BE OF THE SIZES: MAX_ZONE, MAX_NETWORK, MAX_HOST!!!
	This is for bufferoverflow prevention.

*/
int
validate_zonename(int debuglvl, char *zonename, int onlyvalidate, char *zone, char *network, char *host, regex_t *reg_ex)
{
	int		retval=0;
	// this initalization pleases splint
	regmatch_t	reg_match[8] = {{0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}};

	if(debuglvl >= HIGH) (void)vrprint.debug("Debug", "validate_zonename: checking: %s", zonename);

	if(onlyvalidate == 1)
	{
		if(regexec(reg_ex, zonename, 0, NULL, 0) != 0)
		{
			(void)vrprint.error(-1, "Error", "Zonename '%s' is invalid. A zonename can contain normal letters and numbers and the underscore (_) and minus (-) characters.", zonename);
			retval=-1;
		}
	}

	if(onlyvalidate == 0)
	{
		if(regexec(reg_ex, zonename, 8, reg_match, 0) != 0)
		{
			(void)vrprint.warning("Warning", "Zonename '%s' is invalid.", zonename);
			retval=-1;
		}
		else
		{

			if(onlyvalidate == 0)
			{
				if(reg_match[7].rm_eo-reg_match[7].rm_so == 0)
				{
					host[0]='\0';

					if(reg_match[4].rm_eo-reg_match[4].rm_so == 0)
					{
						network[0]='\0';

						if(reg_match[1].rm_eo-reg_match[1].rm_so == 0)
						{
							zone[0]='\0';
							retval=-1;
						}
						else
						{
							(void)range_strcpy(zone, zonename, reg_match[1].rm_so, reg_match[1].rm_eo, MAX_ZONE);
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "validate_zonename: zone: %s.", zone);
						}
					}
					else
					{
						(void)range_strcpy(network, zonename, reg_match[1].rm_so, reg_match[1].rm_eo, MAX_NETWORK);
						(void)range_strcpy(zone, zonename, reg_match[4].rm_so, reg_match[4].rm_eo, MAX_ZONE);
						if(debuglvl >= HIGH) (void)vrprint.debug("Debug", "validate_zonename: zone: %s, network: %s.", zone, network);
					}
				}
				else
				{
					(void)range_strcpy(host, zonename, reg_match[1].rm_so, reg_match[1].rm_eo, MAX_HOST);
					(void)range_strcpy(network, zonename, reg_match[4].rm_so, reg_match[4].rm_eo, MAX_NETWORK);
					(void)range_strcpy(zone, zonename, reg_match[7].rm_so, reg_match[7].rm_eo, MAX_ZONE);
					if(debuglvl >= HIGH) (void)vrprint.debug("Debug", "validate_zonename: zone: %s, network: %s, host: %s.", zone, network, host);
				}
			}
			else
			{
				if(debuglvl >= MEDIUM) (void)vrprint.debug("Debug", "validate_zonename: '%s' is valid.", zonename);
			}
		}
	}

	return(retval);
}

/*	zones_group_save_members

	Save the group members to the backend.
	
	Returncodes:
		 0: ok
		-1: error
*/
int
zones_group_save_members(const int debuglvl, struct ZoneData_ *group_ptr)
{
	d_list_node		*d_node = NULL;
	struct ZoneData_	*member_ptr = NULL;
	
	/* safety */
	if(!group_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* safety */
	if(group_ptr->GroupList.len < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "negative number of groupmembers (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* write to backend */
	if(group_ptr->GroupList.len == 0)
	{
		/* clear */
		if(zf->tell(debuglvl, zone_backend, group_ptr->name, "MEMBER", "", 1, TYPE_GROUP) < 0)
		{
			(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
			return(-1);
		}
	}
	else
	{
		/* write to backend */
		for(d_node = group_ptr->GroupList.top; d_node; d_node = d_node->next)
		{
			if(!(member_ptr = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
				return(-1);
			}

			if(d_node == group_ptr->GroupList.top)
			{
				/* save to backend */
				if(zf->tell(debuglvl, zone_backend, group_ptr->name, "MEMBER", member_ptr->host_name, 1, TYPE_GROUP) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
			else
			{
				/* save to backend */
				if(zf->tell(debuglvl, zone_backend, group_ptr->name, "MEMBER", member_ptr->host_name, 0, TYPE_GROUP) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
		}
	}
	
	return(0);
}

/*	zones_network_get_interfaces

	Returncodes:
		 0: ok
		-1: error
 */
int
zones_network_get_interfaces(const int debuglvl, struct ZoneData_ *zone_ptr, Interfaces *interfaces)
{
	int 			result = 0,
				at_least_one_active = 0;
	char			cur_ifac[MAX_INTERFACE] = "";
	struct InterfaceData_	*iface_ptr = NULL;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** start **", __FUNCTION__);

	/* safety */
	if(!zone_ptr || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check if the zone is a network */
	if(zone_ptr->type != TYPE_NETWORK)
	{
		(void)vrprint.error(-1, "Internal Error", "zone '%s' is not a network, but a '%d' (in: %s).", zone_ptr->name, zone_ptr->type, __FUNCTION__);
		return(-1);
	}

	/* reset active interfaces */
	zone_ptr->active_interfaces = 0;

	/* get all interfaces from the backend */
	while((result = zf->ask(debuglvl, zone_backend, zone_ptr->name, "INTERFACE", cur_ifac, sizeof(cur_ifac), TYPE_NETWORK, 1)) == 1)
	{
		/* search the interface in the interface list */
		if(!(iface_ptr = search_interface(debuglvl, interfaces, cur_ifac)))
		{
			(void)vrprint.warning("Warning", "the interface '%s' of network '%s' was not found in memory.", cur_ifac, zone_ptr->name);
		}
		else
		{
			if(iface_ptr->active == 0)
			{
				(void)vrprint.warning("Warning", "The interface '%s' of network '%s' is not active.", cur_ifac, zone_ptr->name);
			}
			else
			{
				/* count the active interfaces */
				zone_ptr->active_interfaces++;
				at_least_one_active = 1;
			}

			/* append to the list */
			if(!(d_list_append(debuglvl, &zone_ptr->InterfaceList, iface_ptr)))
			{
				(void)vrprint.error(-1, "Internal Error", "appending to the list failed (in: %s).", __FUNCTION__);
				return(-1);
			}
		}
	}

	/* set the interface count to the list length */
	zone_ptr->interface_count = zone_ptr->InterfaceList.len;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: interface_count: %d, active_interfaces: %d.", __FUNCTION__, zone_ptr->interface_count, zone_ptr->active_interfaces);

	/* if we don't have at least one active interface print a warning and deactivate */
	if(at_least_one_active == 0)
	{
		(void)vrprint.warning("Warning", "Network '%s' doesn't have at least one active interface, deactivating network.", zone_ptr->name);
		zone_ptr->active = 0;
	}

	return(0);
}


int
zones_network_save_interfaces(const int debuglvl, struct ZoneData_ *network_ptr)
{
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;
	
	/* safety */
	if(!network_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check if the zone is a network */
	if(network_ptr->type != TYPE_NETWORK)
	{
		(void)vrprint.error(-1, "Internal Error", "zone '%s' is not a network, but a '%d' (in: %s).", network_ptr->name, network_ptr->type, __FUNCTION__);
		return(-1);
	}

	/* safety */
	if(network_ptr->InterfaceList.len < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "negative number of interfacemembers (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* write the new list to the backend */
	if(network_ptr->InterfaceList.len == 0)
	{
		/* clear by writing "" in overwrite mode */
		if(zf->tell(debuglvl, zone_backend, network_ptr->name, "INTERFACE", "", 1, TYPE_NETWORK) < 0)
		{
			(void)vrprint.error(-1, "Error", "writing to backend failed (in: %s).", __FUNCTION__);
			return(-1);
		}
	}
	else
	{
		/*
			save the new interfaces list to the backend
		*/
		for(d_node = network_ptr->InterfaceList.top; d_node; d_node = d_node->next)
		{
			if(!(iface_ptr = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
				return(-1);
			}

			if(d_node == network_ptr->InterfaceList.top)
			{
				/* the first one is in overwrite mode */
				if(zf->tell(debuglvl, zone_backend, network_ptr->name, "INTERFACE", iface_ptr->name, 1, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "writing to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
			else
			{
				/* no overwriting, just appending */
				if(zf->tell(debuglvl, zone_backend, network_ptr->name, "INTERFACE", iface_ptr->name, 0, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "writing to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
		}
	}
	
	return(0);
}


/*	rules_parse_line

	Returncodes:
		0: ok
		-1: error
*/
static int
zones_network_protectrule_parse_line(const int debuglvl, const char *line, struct RuleData_ *rule_ptr)
{
	int	line_pos = 0,	// position in line
		var_pos=0;	// position in varible
	char	against_keyw[32] = "";


	/* safety first */
	if(!line || !rule_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* this should not happen, but it can't hurt to check, right? */
	if(strlen(line) > MAX_RULE_LENGTH)
	{
		(void)vrprint.error(-1, "Internal Error", "rule is too long (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* get the action */
	for(; line_pos < sizeof(rule_ptr->action)-1 && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		rule_ptr->action[var_pos] = line[line_pos];
	}
	rule_ptr->action[var_pos] = '\0';

	/* now we analyze the action */
	if(strcasecmp(rule_ptr->action, "protect") != 0)
	{
		(void)vrprint.error(-1, "Error", "expected action protect, got '%s' (in: %s:%d).", rule_ptr->action, __FUNCTION__, __LINE__);
		return(-1);
	}

	/* get the 'against' */
	for(line_pos++, var_pos = 0; var_pos < sizeof(against_keyw) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		against_keyw[var_pos] = line[line_pos];
	}
	against_keyw[var_pos] = '\0';

	/* check for the against keyword */
	if(strcasecmp(against_keyw, "against") != 0)
	{
		(void)vrprint.error(-1, "Error", "expected action protect, got '%s' (in: %s:%d).", rule_ptr->action, __FUNCTION__, __LINE__);
		return(-1);
	}

	/* okay, now lets see what kind of danger we are talking about */
	for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.danger) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		rule_ptr->danger[var_pos] = line[line_pos];
	}
	rule_ptr->danger[var_pos] = '\0';

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: protect: danger: '%s'", __FUNCTION__, rule_ptr->danger);

	/* now determine if the danger is 'spoofing' */
	if(strcasecmp(rule_ptr->danger, "spoofing") != 0)
	{
		(void)vrprint.error(-1, "Error", "expected action protect, got '%s' (in: %s:%d).", rule_ptr->action, __FUNCTION__, __LINE__);
		return(-1);
	}

	/* get the 'from' */
	for(line_pos++, var_pos = 0; var_pos < strlen("from") && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		rule_ptr->source[var_pos] = line[line_pos];
	}
	rule_ptr->source[var_pos] = '\0';

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: protect: keyword from: '%s'", __FUNCTION__, rule_ptr->source);

	/* if 'from' is missing, the rule is malformed, so we bail out screaming & kicking */
	if(strcasecmp(rule_ptr->source, "from") != 0)
	{
		(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'from' is missing: %s (in: %s).", line, __FUNCTION__);
		return(-1);
	}

	/* get the source */
	for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.source) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		rule_ptr->source[var_pos] = line[line_pos];
	}
	rule_ptr->source[var_pos] = '\0';

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: protect: source: '%s'", __FUNCTION__, rule_ptr->source);

	/* set the ruletype */
	rule_ptr->type = PROT_IPTABLES;

	return(0);
}


int
zones_network_get_protectrules(const int debuglvl, struct ZoneData_ *network_ptr)
{
	int 			result = 0;
	char			currule[MAX_RULE_LENGTH] = "";
	struct RuleData_	*rule_ptr = NULL;
	d_list_node		*d_node = NULL;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** start **", __FUNCTION__);

	/* safety */
	if(!network_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check if the zone is a network */
	if(network_ptr->type != TYPE_NETWORK)
	{
		(void)vrprint.error(-1, "Internal Error", "zone '%s' is not a network, but a '%d' (in: %s).", network_ptr->name, network_ptr->type, __FUNCTION__);
		return(-1);
	}

	/* get all protect rules from the backend */
	while((result = zf->ask(debuglvl, zone_backend, network_ptr->name, "RULE", currule, sizeof(currule), TYPE_NETWORK, 1)) == 1)
	{
		/* get mem */
		if(!(rule_ptr = rule_malloc()))
			return(-1);

		/* copy name */
		if(strlcpy(rule_ptr->who, network_ptr->name, sizeof(rule_ptr->who)) >= sizeof(rule_ptr->who))
		{
			(void)vrprint.error(-1, "Internal Error", "buffer too small (in: %s:%d).", __FUNCTION__, __LINE__);
			free(rule_ptr);
			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "currule: '%s'.", currule);

		if(zones_network_protectrule_parse_line(debuglvl, currule, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "parsing protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			free(rule_ptr);
			return(-1);
		}

		/* append to list */
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) == NULL)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule to list failed (in: %s:%d).", __FUNCTION__, __LINE__);
			free(rule_ptr);
			return(-1);
		}
	}

	for(d_node = network_ptr->ProtectList.top; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "a: %s, w: %s, d: %s, s: %s.", rule_ptr->action, rule_ptr->who, rule_ptr->danger, rule_ptr->source);
	}

	return(0);
}
