/***************************************************************************
 *   Copyright (C) 2002-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/***************************************************************************
 * Here we try to create the actual rule.                                  *
 ***************************************************************************/
#include "main.h"


int
analyze_interface_protect_rules(const int debuglvl, Rules *rules, Zones *zones, Services *services, Interfaces *interfaces)
{
	struct RuleData_	*rule_ptr = NULL;
	d_list_node		*d_node = NULL,
				*if_d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;

	/* safety */
	if(!rules || !zones || !services || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	/* first analyze the protectrules in the interfaces */
	for(d_node = interfaces->list.top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		for(if_d_node = iface_ptr->ProtectList.top; if_d_node; if_d_node = if_d_node->next)
		{
			if(!(rule_ptr = if_d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			if(rules_analyze_rule(debuglvl, rule_ptr, &rule_ptr->rulecache, services, zones, interfaces) == 0)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug(__FUNCTION__, "analizing protectrule success, active = 1.");

				rule_ptr->active = 1;
			}
			else
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug(__FUNCTION__, "analizing protectrule failed, active = 0.");

				rule_ptr->active = 0;
			}
		}
	}

	return(0);
}


int
analyze_network_protect_rules(const int debuglvl, Rules *rules, Zones *zones, Services *services, Interfaces *interfaces)
{
	struct RuleData_	*rule_ptr = NULL;
	d_list_node		*d_node = NULL,
				*net_d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;

	/* safety */
	if(!rules || !zones || !services || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	/* first analyze the protectrules in the network */
	for(d_node = zones->list.top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		if(zone_ptr->type == TYPE_NETWORK)
		{
			for(net_d_node = zone_ptr->ProtectList.top; net_d_node; net_d_node = net_d_node->next)
			{
				if(!(rule_ptr = net_d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				if(rules_analyze_rule(debuglvl, rule_ptr, &rule_ptr->rulecache, services, zones, interfaces) == 0)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug(__FUNCTION__, "analizing protectrule success, active = 1.");

					rule_ptr->active = 1;
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug(__FUNCTION__, "analizing protectrule failed, active = 0.");

					rule_ptr->active = 0;
				}
			}
		}
	}

	return(0);
}


int
analyze_normal_rules(const int debuglvl, Rules *rules, Zones *zones, Services *services, Interfaces *interfaces)
{
	struct RuleData_	*rule_ptr = NULL;
	int			rulescount = 0,
				rulesfailedcount = 0;
	d_list_node		*d_node = NULL,
				*next_d_node = NULL;

	/* safety */
	if(!rules || !zones || !services || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	/* check if the list is not empty. If it is, d_node will be NULL. */
	if(rules->list.len > 0)
	{
		/*	Get the top of the list. This should never fail because
			we already checked the listsize.
		*/
		if(!(d_node = rules->list.top))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	/* if we have a node, continue */
	for(;d_node;)
	{
		rulescount++;

		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* analyze! */
		if(rules_analyze_rule(debuglvl, rule_ptr, &rule_ptr->rulecache, services, zones, interfaces) == 0)
		{
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug("Debug", "main: analyze_rule %3d ok.", rulescount);
				
			/* update d_node */
			d_node = d_node->next;
		}
		else
		{
			(void)vrprint.warning("Warning", "Analyzing rule %d failed.", rulescount);
			rulesfailedcount++;
				
			/* update node before removing */
			next_d_node = d_node->next;

			/* remove the failed rule from the list */
			if(d_list_remove_node(debuglvl, &rules->list, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "removing node from list failed (in: analyze_all_rules).");
				return(-1);
			}

			/* now update the number in the list */
			rules_update_numbers(debuglvl, rules, rulescount - rulesfailedcount + 1, 0);

			free_options(debuglvl, rule_ptr->opt);
			free(rule_ptr);
			rule_ptr = NULL;
				
			/* set d_node to the next_d_node */
			d_node = next_d_node;
		}
	}

	(void)vrprint.info("Info", "Analyzing %d rules succesfull, %d failed.", rulescount - rulesfailedcount, rulesfailedcount);
	return(0);
}


/*	analyze_all_rules

	Analyzes all rules :-)

	Returncodes:
		 0: ok
		-1: error
*/
int
analyze_all_rules(const int debuglvl, Rules *rules, Zones *zones, Services *services, Interfaces *interfaces)
{
	/* safety */
	if(!rules || !zones || !services || !interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	(void)vrprint.info("Info", "Analyzing the rules... ");

	/* interface rules */
	if(analyze_interface_protect_rules(debuglvl, rules, zones, services, interfaces) < 0)
		return(-1);

	/* network rules */
	if(analyze_network_protect_rules(debuglvl, rules, zones, services, interfaces) < 0)
		return(-1);

	/* normal rules */
	if(analyze_normal_rules(debuglvl, rules, zones, services, interfaces) < 0)
		return(-1);

	return(0);
}


/*	create_all_rules

	Creates all rules.

	If 'create_prerules' is set to 1, prerules() will be called.

	Returncodes:
		 0: ok
		-1: error
*/
int
create_all_rules(const int debuglvl, Rules *rules, Zones *zones, Interfaces *interfaces, d_list *block_list, IptCap *iptcap, int create_prerules)
{
	int			result = 0;
	char			forward_rules = 0;

	/* safety */
	if(!rules || !interfaces || !block_list || !iptcap || create_prerules < 0 || create_prerules > 1)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* loglevel */
	if(strcmp(conf.loglevel, "") == 0)
	{
		memset(loglevel, 0, sizeof(loglevel));
	}
	else
	{
		/* create the loglevel string */
		if(snprintf(loglevel, sizeof(loglevel), "--log-level %s", conf.loglevel) >= sizeof(loglevel))
		{
			(void)vrprint.error(-1, "Error", "buffer overrun (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	(void)vrprint.info("Info", "Creating the rules... (rules to create: %d)", rules->list.len);

	/* create the prerules if were called with it */
	if(create_prerules)
	{
		result = pre_rules(debuglvl, NULL, interfaces, iptcap);
		if(result < 0)
			return(-1);
	}

	/* create the blocklist */
	if(conf.use_blocklist)
	{
		if(create_block_rules(debuglvl, NULL, block_list) < 0)
		{
			(void)vrprint.error(-1, "Error", "create blocklist failed.");
		}
	}

	/* create the network protect rules (anti-spoofing) */
	if(create_network_protect_rules(debuglvl, NULL, zones, iptcap) < 0)
	{
		(void)vrprint.error(-1, "Error", "create protectrules failed.");
	}
	/* create the interface protect rules (proc stuff) */
	if(create_interface_protect_rules(debuglvl, interfaces) < 0)
	{
		(void)vrprint.error(-1, "Error", "create protectrules failed.");
	}
	/* system protect rules (proc) */
	if(create_system_protectrules(debuglvl, &conf) < 0)
	{
		(void)vrprint.error(-1, "Error", "create protectrules failed.");
	}
	/* normal rules, ruleset == NULL */
	if(create_normal_rules(debuglvl, NULL, rules, iptcap, &forward_rules) < 0)
	{
		(void)vrprint.error(-1, "Error", "create normal rules failed.");
	}

	/* post rules: enable logging */
	if(post_rules(debuglvl, NULL, iptcap, forward_rules) < 0)
		return(-1);

	(void)vrprint.info("Info", "Creating rules finished.");
	return(0);
}


static int
create_rule_set_ports(struct RuleCreateData_ *rule, struct portdata *portrange_ptr)
{
	/* from */
	if(portrange_ptr->protocol == 6 || portrange_ptr->protocol == 17)
	{
		if(portrange_ptr->src_high == 0)
			snprintf(rule->temp_src_port, sizeof(rule->temp_src_port), "--sport %d", portrange_ptr->src_low);
		else
			snprintf(rule->temp_src_port, sizeof(rule->temp_src_port), "--sport %d:%d", portrange_ptr->src_low, portrange_ptr->src_high);
	}
	else if(portrange_ptr->protocol == 1 || portrange_ptr->protocol == 47 || portrange_ptr->protocol == 50 || portrange_ptr->protocol == 51)
		snprintf(rule->temp_src_port, sizeof(rule->temp_src_port), " ");

	/* to */
	if(portrange_ptr->protocol == 6 || portrange_ptr->protocol == 17)
	{
		if(portrange_ptr->dst_high == 0)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", portrange_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", portrange_ptr->dst_low, portrange_ptr->dst_high);
	}
	else if(portrange_ptr->protocol == 1)
	{
		if(portrange_ptr->dst_high == -1)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--icmp-type %d", portrange_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--icmp-type %d/%d", portrange_ptr->dst_low, portrange_ptr->dst_high);
	}
	else if(portrange_ptr->protocol == 47 || portrange_ptr->protocol == 50 || portrange_ptr->protocol == 51)
		snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), " ");

	return(0);
}


static int
create_rule_set_proto(struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(rule->portrange_ptr)
	{
		/* tcp */
		if(rule->portrange_ptr->protocol == 6)
			strlcpy(rule->proto, "-p tcp --syn", sizeof(rule->proto));
		/* udp */
		else if(rule->portrange_ptr->protocol == 17)
			strlcpy(rule->proto, "-p udp", sizeof(rule->proto));
		/* icmp */
		else if(rule->portrange_ptr->protocol == 1)
			strlcpy(rule->proto, "-p icmp", sizeof(rule->proto));
		/* gre */
		else if(rule->portrange_ptr->protocol == 47)
			strlcpy(rule->proto, "-p gre", sizeof(rule->proto));
		else
		{
			if(rule->portrange_ptr->protocol > 0 && rule->portrange_ptr->protocol <= 255)
				snprintf(rule->proto, sizeof(rule->proto), "-p %d", rule->portrange_ptr->protocol);
			else
				strlcpy(rule->proto, "", sizeof(rule->proto));
		}
	}
	
	/* if we have no protocol (usefull for service 'all') */
	if(	(create->service->count == 1) &&
		
		(create->service->tcp_count +
		create->service->udp_count +
		create->service->icmp_count +
		create->service->gre_count +
		create->service->esp_count +
		create->service->ah_count == 0))
	{
		strcpy(rule->proto, "");
		strcpy(rule->temp_src_port, "");
		strcpy(rule->temp_dst_port, "");
	}

	return(0);
}


static int
create_rule_set_action_and_ip(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create, IptCap *iptcap)
{
	char	action[9] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/*	clear rule->limit because we only use it with log rules and if loglimit > 0
		and if iptables has the capability
	*/
	memset(rule->limit, 0, sizeof(rule->limit));

	/* get the first part of the action, because action can be like this: REJECT --reject-type icmp-adm.... */
	sscanf(create->action, "%s", action);

	/* if we want to log a rule, but havent done it yet: */
	if(create->option.rule_log && !rule->logged)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: log the rule.", __FUNCTION__);

		/* create the limitstring */
		if(iptcap->match_limit || bash_out)
		{
			/* set the loglimit */
			if(create->option.loglimit > 0)
				snprintf(rule->limit, sizeof(rule->limit), "-m limit --limit %d/s --limit-burst %d",
										create->option.loglimit,
										create->option.logburst);
		}

		/* with prefix */
		if(create->option.rule_logprefix == 1)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "log with user prefix.");

			/* see if we use a loglevel */
			snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix \"vrmr: %s %s \" %s", action, create->option.logprefix, loglevel);
		}
		else
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "log without user prefix.");

			/* see if we use a loglevel */
			snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix \"vrmr: %s \" %s", action, loglevel);
		}

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.ipaddress, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, rule->ipv4_to.netmask,   sizeof(rule->to_netmask));

		/* tag logged done */
		rule->logged = 1;
	}
	/* now create the normal rule */
	else if(!rule->normal)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "create the normal rule.");

		/* action LOG requires some extra attention */
		if(strncasecmp(create->action, "LOG", 3) == 0)
		{
			if(iptcap->match_limit || bash_out)
			{
				/* set the loglimit */
				if(create->option.loglimit > 0)
				{
					snprintf(rule->limit, sizeof(rule->limit), "-m limit --limit %d/s --limit-burst %d",
											create->option.loglimit,
											create->option.logburst);
				}
			}

			snprintf(rule->action, sizeof(rule->action), "%s --log-prefix \"vrmr: %s %s \"",
									create->action,
									action,
									create->option.logprefix);
		}
		else
		{
			strlcpy(rule->action, create->action, sizeof(rule->action));
		}

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.ipaddress, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, rule->ipv4_to.netmask,   sizeof(rule->to_netmask));
		
		/* tag normal rule done */
		rule->normal = 1;
	}
	/* if we have a broadcasting protocol and want logging, and haven't logged yet */
	else if((create->service->broadcast == 1) && (create->option.rule_log == 1) && (rule->broadlog == 0))
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "create the log rule for broadcast.");

		if(iptcap->match_limit || bash_out)
		{
			/* set the loglimit */
			if(create->option.loglimit > 0)
				snprintf(rule->limit, sizeof(rule->limit), "-m limit --limit %d/s --limit-burst %d",
										create->option.loglimit,
										create->option.logburst);
		}

		if(create->option.rule_logprefix == 1)
		{
			snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix \"vrmr: %s %s \" %s",
									create->action,
									create->option.logprefix,
									loglevel);
		}
		else
		{
			if(strcmp(loglevel, "") == 0)
				strlcpy(rule->action, "LOG", sizeof(rule->action));
			else
				snprintf(rule->action, sizeof(rule->action), "LOG %s", loglevel);
		}

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.broadcast, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, "255.255.255.255", sizeof(rule->to_netmask));

		/* tag broadcast logging done */
		rule->broadlog = 1;
	}
	/* broadcasting */
	else if((create->service->broadcast == 1) && (rule->broadcasted == 0))
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "create the broadcast rule.");

		strlcpy(rule->action, create->action, sizeof(rule->action));

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.broadcast, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, "255.255.255.255", sizeof(rule->to_netmask));

		/* tag broadcasting done */
		rule->broadcasted = 1;
	}

	/* all the above stuff is only for 'to'. */
	strlcpy(rule->from_ip,      rule->ipv4_from.ipaddress, sizeof(rule->from_ip));
	strlcpy(rule->from_netmask, rule->ipv4_from.netmask, sizeof(rule->from_netmask));

	return(0);
}


/*	create_rule

	This fuctions creates the actual rule.

	Returncodes:
		 0: ok
		-1: error
*/
int
create_rule(const int debuglvl, /*@null@*/RuleSet *ruleset, struct RuleCache_ *create, IptCap *iptcap)
{
	int			retval = 0;
	int			z = 0;
	int			from_loop_cnt = 0,
				to_loop_cnt = 0,
				from_nor,
				to_nor;
	int			linescreated = 0,
				createloop = 0;

	struct RuleCreateData_ 	*rule = NULL;

	char			from_mac[19] = "";
	int			from_has_mac = 0;

	struct InterfaceData_	*from_if_ptr = NULL,
				*to_if_ptr = NULL;

	d_list_node		*from_d_node = NULL,
				*to_d_node = NULL;

	d_list_node		*from_if_node = NULL,
				*to_if_node = NULL;

	struct ZoneData_ 	*from_zone_ptr = NULL,
				*to_zone_ptr = NULL,
				*from_member_ptr = NULL,
				*to_member_ptr = NULL;

	int 			from_active = 0,
				to_active = 0,
				active = 0;

	d_list_node		*port_d_node = NULL,
				*listenport_d_node = NULL,
				*remoteport_d_node = NULL;

	struct portdata		*portrange_ptr = NULL;


	/* safety checks */
	if(!iptcap || !create || !create->from || !create->to || !create->service)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** start ** (create->action: %s).", __FUNCTION__, create->action);

	/* here we print the description if we are in bashmode */
	if(bash_out == 1 && create->description != NULL)
	{
		fprintf(stdout, "\n# %s\n", create->description);
		
		free(create->description);
		create->description = NULL;
	}

	/* clear counters */
	create->iptcount.input = 0,
	create->iptcount.output = 0,
	create->iptcount.forward = 0,
	create->iptcount.preroute = 0,
	create->iptcount.postroute = 0;

	/* at start we asume that both from and to are active */
	from_active = 1;
	to_active = 1;
	active = create->active;

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "create_rule: count: %d, tcp_count: %d, udp_count: %d, icmp_count: %d.", create->service->count, create->service->tcp_count, create->service->udp_count, create->service->icmp_count);

	/* if bash, print comment (if any) */
	if(create->option.rule_comment && bash_out)
	{
		fprintf(stdout, "# comment: '%s'\n", create->option.comment);
	}

	/* alloc the temp rule data */
	if(!(rule = malloc(sizeof(struct RuleCreateData_))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* init */
	memset(rule, 0, sizeof(struct RuleCreateData_));

	/* copy the helper */
	strlcpy(rule->helper, create->service->helper, sizeof(rule->helper));

////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// LOOP CONTROL /////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

	/* FROM zone, network, host or group */
	if(create->from->type == TYPE_GROUP)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: groupmembercount: from: %d.", __FUNCTION__, create->from->group_member_count);

		/* set the from-loop counter to the group-member-count */
		from_loop_cnt = create->from->group_member_count;

		/* set the first member */
		from_d_node = create->from->GroupList.top;

		/* see if it is not NULL */
		if(from_d_node)
		{
			if(!(from_member_ptr = from_d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			/* active check */
			if(!from_member_ptr->active)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "%s: groupmember '%s' is inactive.", __FUNCTION__, from_member_ptr->name);

				from_active = 0;
			}
			else
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "%s: groupmember '%s' is active.", __FUNCTION__, from_member_ptr->name);

				from_active = 1;
			}
		}
		/* safety check */
		else if(!from_d_node && from_loop_cnt > 0)
		{
			(void)vrprint.error(-1, "Internal Error", "from_loop_cnt: %d, but from_d_node == NULL for group %s (in: %s).", from_loop_cnt, create->from->name, __FUNCTION__);
			return(-1);
		}
	}
	/* not a group */
	else
	{
		/* just one from_loop_cnt */
		from_loop_cnt = 1;
	}

	/* TO zone, network, host or group */
	if(create->to->type == TYPE_GROUP)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: groupmembercount: to: %d.", __FUNCTION__, create->to->group_member_count);

		/* set the to-loop counter to the group-member-count */
		to_loop_cnt = create->to->group_member_count;

		/* set the first member */
		to_d_node = create->to->GroupList.top;

		/* see if it is not NULL */
		if(to_d_node)
		{
			if(!(to_member_ptr = to_d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
				return(-1);
			}

			if(!to_member_ptr->active)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "%s: groupmember '%s' is inactive.", __FUNCTION__, to_member_ptr->name);

				to_active = 0;
			}
			else
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "%s: groupmember '%s' is active.", __FUNCTION__, to_member_ptr->name);

				to_active = 1;
			}
		}
		/* safety check */
		else if(!to_d_node && to_loop_cnt > 0)
		{
			(void)vrprint.error(-1, "Internal Error", "to_loop_cnt: %d, but to_d_node == NULL for group %s (in: %s).", to_loop_cnt, create->to->name, __FUNCTION__);
			return(-1);
		}
	}
	/* not a group */
	else
	{
		/* to_loop_cnt is one if it's not a group */
		to_loop_cnt = 1;
	}

	/*	determine how many times we are going to run the next loops

		multiply by the number of interfaces - we need to make sure that interfaces are not counted twice */

	/* from zone */
	if(create->from->type == TYPE_HOST || create->from->type == TYPE_GROUP)
	{
		/* check for counting interfaces twice */
		if((create->to->type == TYPE_NETWORK) && (create->to == create->from->network_parent))
		{
			/* do nothing */
		}
		else
		{
			from_loop_cnt = from_loop_cnt * create->from->network_parent->interface_count;
		}
	}
	else if(create->from->type == TYPE_NETWORK)
	{
		from_loop_cnt = from_loop_cnt * create->from->interface_count;
	}
	else
	{
		(void)vrprint.error(-1, "Internal Error", "invalid from type %d (in: %s).", create->from->type, __FUNCTION__);
		return(-1);
	}

	/* to zone */
	if(create->to->type == TYPE_HOST || create->to->type == TYPE_GROUP)
	{
		/* check for counting twice: compare from parentnetwork with our to parentnetwork */
		if((create->from->type == TYPE_HOST || create->from->type == TYPE_GROUP) && create->from->network_parent == create->to->network_parent)
		{
			/* do nothing */
		}
		/* check for twice:  compare from network with our parentnetwork */
		else if(create->from->type == TYPE_NETWORK && create->from == create->to->network_parent)
		{
			/* do nothing */
		}
		/* check for counting twice: else multiply */
		else
		{
			to_loop_cnt = to_loop_cnt * create->to->network_parent->interface_count;
		}
	}
	else if(create->to->type == TYPE_NETWORK)
	{
		/* check for counting twice: compare from network with our network */
		if(create->from->type == TYPE_NETWORK && create->from == create->to)
		{
			/* do nothing */
		}
		/* check for counting twice:  else multiply */
		else
		{
			to_loop_cnt = to_loop_cnt * create->to->interface_count;
		}
	}
	else
	{
		(void)vrprint.error(-1, "Internal Error", "invalid to type %d (in: %s).", create->from->type, __FUNCTION__);
		return(-1);
	}


	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "create_rule: loop: from_loop_cnt: %d, to_loop_cnt: %d.", from_loop_cnt, to_loop_cnt);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////// MAIN LOOP ///////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////
// BEGIN LOOP INTERFACES
////

	/*
	 * ==== FROM LOOP ====
	 */
	/* run this loop 'from' times */
	for(from_nor = 0;  from_nor < from_loop_cnt; from_nor++)
	{
		/* network or host, group */
		if(create->from->type == TYPE_NETWORK)
			from_zone_ptr = create->from;
		else
			from_zone_ptr = create->from->network_parent;

		if(!from_zone_ptr)
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* get the ifac ptr */
		if(!from_if_node)
			from_if_node = from_zone_ptr->InterfaceList.top;
		else
			from_if_node = from_if_node->next;

		/* safety check */
		if(!from_if_node)
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return -1;
		}

		/* reset from_zone_ptr */
		from_zone_ptr = NULL;

		if(!(from_if_ptr = from_if_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/*	set the interface

			if the device is virtual (oldstyle) we don't want it in our
			iptables commands
		*/
		if(!from_if_ptr->device_virtual_oldstyle)
			strlcpy(rule->from_int, from_if_ptr->device, sizeof(rule->from_int));
		else
			bzero(rule->from_int, sizeof(rule->from_int));

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "create_rule: rule->from_int: '%s', virtual? %s.", rule->from_int, from_if_ptr->device_virtual_oldstyle ? "Yes" : "No");

		/*	if the interface is not active we set the to zone
			inactive. But only if its already active. */
		if(active)
			active = from_if_ptr->active;

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "create_rule: active = %d (because of from_if_ptr->active).", active);

		/* if were at the end of the ifac list, set to null */
		if(d_list_node_is_bot(debuglvl, from_if_node))
			from_if_node = NULL;

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "create_rule: create->from.group_member_count: %d.", create->from->group_member_count);

		/*	check for the 'int' option:
 
		 	3 possibilities:
 
		 	1. portfw or redirect and interface option set
		 	2. portfw or redirect and interface option not set
		 	3. not portfw or redirect
		 */
		if(	((create->ruletype == RT_PORTFW || create->ruletype == RT_REDIRECT) &&	/* portfw or redirect */
			create->option.interface[0] != '\0' &&					/* interface option is set */
			strcmp(create->option.interface, from_if_ptr->name) == 0)		/* interface matches */
				||
			((create->ruletype == RT_PORTFW || create->ruletype == RT_REDIRECT) &&	/* portfw or redirect */
			create->option.interface[0] == '\0')					/* interface option is not set */
				||
			(create->ruletype != RT_PORTFW && create->ruletype != RT_REDIRECT)	/* or not portfw and redirect */
		)
		{
			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNCTION__, "going on to create... (no portfw or redirect, int option not set or int option matched).");

		/*
		 * ==== TO LOOP ====
		 */
			/* run this loop 'to' times */
			for(to_nor = 0; to_nor < to_loop_cnt; to_nor++)
			{
				/* set to zone pointer to the network of this zone */
				if(create->to->type == TYPE_NETWORK)
					to_zone_ptr = create->to;
				else
					to_zone_ptr = create->to->network_parent;

				/* check */
				if(!to_zone_ptr)
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				/* get the ifac node */
				if(!to_if_node)
					to_if_node = to_zone_ptr->InterfaceList.top;
				else
					to_if_node = to_if_node->next;

				/* safety check */
				if(!to_if_node)
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return -1;
				}

				/* reset to_zone_ptr */
				to_zone_ptr = NULL;

				/* get the interface */
				if(!(to_if_ptr = to_if_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				/*	set the interface

					if the device is virtual (oldstyle) we don't want it in our
					iptables commands
				*/
				if(!to_if_ptr->device_virtual_oldstyle)
					strlcpy(rule->to_int, to_if_ptr->device, sizeof(rule->to_int));
				else
					bzero(rule->to_int, sizeof(rule->to_int));

				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "create_rule: rule->to_int: '%s', virtual? %s.", rule->to_int, to_if_ptr->device_virtual_oldstyle ? "Yes" : "No");

				/* if the interface is not active we set the rule inactive. */
				if(active)
				{
					active = to_if_ptr->active;

					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "%s: active = %d (because of to_if_ptr->active).", __FUNCTION__, active);
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "%s: active = %d (so not checking the to_int).", __FUNCTION__, active);
				}

				/* if were at the end of the ifac list, set to null so we can rewind at the next run */
				if(d_list_node_is_bot(debuglvl, to_if_node))
					to_if_node = NULL;


			/*
			 * ==== ACTIVE CHECK ====
			 */
				if(from_active && to_active && active)
				{
					/* SET PORTS */
					if(create->service->PortrangeList.len > 0)
					{
						/* top of the list */
						if(!(port_d_node = create->service->PortrangeList.top))
						{
							(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
							return -1;
						}

						/* listenport option */
						if(create->option.listenport == 1)
							listenport_d_node = create->option.ListenportList.top;
						else
							listenport_d_node = NULL;

						/* remoteport option */
						if(create->option.remoteport == 1)
							remoteport_d_node = create->option.RemoteportList.top;
						else
							remoteport_d_node = NULL;
					}
					else
					{
						/* if there are no portranges set port_d_node to NULL */
						port_d_node = NULL;
					}


					for(linescreated = 0; create->service->count > linescreated; linescreated++)
					{
						/* start with creating 1 rule */
						createloop = 1;

						/* set ports */
						if(port_d_node)
						{
							/* get the current portrange */
							if(!(portrange_ptr = port_d_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
								return -1;
							}
							/* set portrange_ptr into rule */
							rule->portrange_ptr = portrange_ptr;

							/* set rule->listenport_ptr */
							if(create->option.listenport && listenport_d_node)
								rule->listenport_ptr = listenport_d_node->data;
							else
								rule->listenport_ptr = NULL;

							/* set rule->remoteport_ptr */
							if(create->option.remoteport && remoteport_d_node)
								rule->remoteport_ptr = remoteport_d_node->data;
							else
								rule->remoteport_ptr = NULL;

							/* now load the ports to the rule struct */
							if(create_rule_set_ports(rule, portrange_ptr) < 0)
							{
								(void)vrprint.error(-1, "Internal Error", "setting up the ports failed (in: %s).", __FUNCTION__);
								retval = -1;
							}
						}
						else
						{
							/* no portrange */
							rule->portrange_ptr = NULL;
						}

						if(debuglvl >= MEDIUM)
						{
							(void)vrprint.debug("Debug", "create_rule: create->service.count: %d, linescreated: %d", create->service->count, linescreated);
							(void)vrprint.debug("Debug", "create_rule: create->service.tcp_count: %d, create->service.udp_count: %d, create->service.icmp_count: %d, create->service.gre_count: %d", create->service->tcp_count, create->service->udp_count, create->service->icmp_count, create->service->gre_count);
						}

						/* set protocol */
						if(create_rule_set_proto(rule, create) < 0)
						{
							(void)vrprint.error(-1, "Error", "create_rule_set_proto failed.");
							retval = -1;
						}

					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					/////////////////////////////// okay, we are going to create something////////////////////////////////////////////////////////
					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

						/*	set the ipaddress and netmask and mac
							
							FROM

							first host
						*/
						from_has_mac = 0;
						if(create->from->type == TYPE_HOST)
						{
							strlcpy(rule->ipv4_from.ipaddress, create->from->ipv4.ipaddress, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   create->from->ipv4.netmask,   sizeof(rule->ipv4_from.netmask));

							if(create->from->has_mac)
							{
								from_has_mac = 1;
								strlcpy(from_mac, create->from->mac, sizeof(from_mac));
							}
						}
						/* network */
						else if(create->from->type == TYPE_NETWORK && create->from_firewall == 0)
						{
							strlcpy(rule->ipv4_from.ipaddress, create->from->ipv4.network, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   create->from->ipv4.netmask, sizeof(rule->ipv4_from.netmask));
						}
						/* firewall (this host) */
						else if(create->from->type == TYPE_NETWORK && create->from_firewall == 1)
						{
							strlcpy(rule->ipv4_from.ipaddress, from_if_ptr->ipv4.ipaddress, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   "255.255.255.255",           sizeof(rule->ipv4_from.netmask));
						}
						/* finally group */
						else if(create->from->type == TYPE_GROUP)
						{
							strlcpy(rule->ipv4_from.ipaddress, from_member_ptr->ipv4.ipaddress, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   from_member_ptr->ipv4.netmask,   sizeof(rule->ipv4_from.netmask));

							if(from_member_ptr->has_mac)
							{
								from_has_mac = 1;
								strlcpy(from_mac, from_member_ptr->mac, sizeof(from_mac));
							}
						}
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "%s: (from) ipaddress: %s, netmask: %s (pre createloop)", __FUNCTION__, rule->ipv4_from.ipaddress, rule->ipv4_from.netmask);

						/*	TO
							
							first host
						*/
						if(create->to->type == TYPE_HOST)
						{
							strlcpy(rule->ipv4_to.ipaddress, create->to->ipv4.ipaddress, sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   create->to->ipv4.netmask,   sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast, sizeof(rule->ipv4_to.broadcast));
							/* to has no mac, because it is not supported by iptables */
						}
						/* network */
						else if(create->to->type == TYPE_NETWORK && create->to_firewall == 0)
						{
							strlcpy(rule->ipv4_to.ipaddress, create->to->ipv4.network,   sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   create->to->ipv4.netmask,   sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast, sizeof(rule->ipv4_to.broadcast));
						}
						/* firewall (this host) */
						else if(create->to->type == TYPE_NETWORK && create->to_firewall == 1)
						{
							strlcpy(rule->ipv4_to.ipaddress, to_if_ptr->ipv4.ipaddress,  sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   "255.255.255.255",          sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast, sizeof(rule->ipv4_to.broadcast));
						}
						else if(create->to->type == TYPE_GROUP)
						{
							strlcpy(rule->ipv4_to.ipaddress, to_member_ptr->ipv4.ipaddress, sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   to_member_ptr->ipv4.netmask,   sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast,    sizeof(rule->ipv4_to.broadcast));
							/* to has no mac, because it is not supported by iptables */
						}
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "%s: ( to ) ipaddress: %s, netmask: %s (pre createloop)", __FUNCTION__, rule->ipv4_to.ipaddress, rule->ipv4_to.netmask);


						/* some special actions for a few actions */
						if(create->ruletype == RT_SNAT)
						{
							/* copy the ipaddress of the to-interface to rule->serverip so snat can use it */
							if(!to_if_ptr)
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer: to_if_ptr (in: %s).", __FUNCTION__);
								return(-1);
							}
								
							snprintf(rule->serverip, sizeof(rule->serverip), "%s", to_if_ptr->ipv4.ipaddress);
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "%s: SNAT: rule->serverip = '%s' (interface: %s).", __FUNCTION__, rule->serverip, to_if_ptr->name);
						}
						else if(create->ruletype == RT_PORTFW)
						{
							/* copy the ipaddress of the from-interface to rule->serverip so portfw can use it */
							if(!from_if_ptr)
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer: from_if_ptr (in: %s).", __FUNCTION__);
								return(-1);
							}
									
							snprintf(rule->serverip, sizeof(rule->serverip), "%s", from_if_ptr->ipv4.ipaddress);
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "%s: PORTFW: rule->serverip = '%s' (interface: %s).", __FUNCTION__, rule->serverip, from_if_ptr->name);
						}


						/* add mac-address if we happen to know it, only 'from' is supported by iptables */
						if(from_has_mac)
						{
							if(iptcap->match_mac || bash_out)
								snprintf(rule->from_mac, sizeof(rule->from_mac), "-m mac --mac-source %s", from_mac);
							else
							{
								(void)vrprint.warning("Warning", "not using macaddress. Mac-match not supported by system.");
								memset(rule->from_mac, 0, sizeof(rule->from_mac));
							}
						}
						else
							memset(rule->from_mac, 0, sizeof(rule->from_mac));

						/* broadcast means twice as many rule */
						if(portrange_ptr && portrange_ptr->protocol == 17 && create->service->broadcast == 1)
							createloop = createloop * 2;

						/* if we log we need to create twice as many rules */
						if(create->option.rule_log)
							createloop = createloop * 2;

						if(debuglvl >= HIGH)
							(void)vrprint.debug(__FUNCTION__, "createloop = %d", createloop);

						/* reset counters */
						rule->logged = 0, rule->normal = 0, rule->broadlog = 0, rule->broadcasted = 0;

					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					////////////////////////////////////////// enter createloop //////////////////////////////////////////////////////
					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
						for(z = 0; z < createloop; z++)
						{
							if(create_rule_set_action_and_ip(debuglvl, rule, create, iptcap) < 0)
							{
								(void)vrprint.error(-1, "Error", "create_rule_set_action_and_ip failed.");
								retval = -1;
							}

							/* rule count */
							ipt_rulecount++;

							if(debuglvl >= HIGH)
							{
								(void)vrprint.debug("Debug", "%s: (from) ipaddress: %s, netmask: %s (in createloop)", __FUNCTION__, rule->from_ip, rule->from_netmask);
								(void)vrprint.debug("Debug", "%s: ( to ) ipaddress: %s, netmask: %s (in createloop)", __FUNCTION__, rule->to_ip, rule->to_netmask);
							}

							/* here we create the iptables commands */
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "%s: action: %s, chain: %s, fromip: %s, toip: %s, ruletype: %d", __FUNCTION__, rule->action, rule->chain, rule->from_ip, rule->to_ip, create->ruletype);

							/* normal input rules */
							if(create->ruletype == RT_INPUT)
							{
								if(create_rule_input(debuglvl, ruleset, rule, create, iptcap) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating input rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* normal output rules */
							else if(create->ruletype == RT_OUTPUT)
							{
								if(create_rule_output(debuglvl, ruleset, rule, create, iptcap) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating output rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* normal forward rules */
							else if(create->ruletype == RT_FORWARD)
							{
								if(create_rule_forward(debuglvl, ruleset, rule, create, iptcap) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating forward rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* masq rules */
							else if(create->ruletype == RT_MASQ)
							{
								if(create_rule_masq(debuglvl, ruleset, rule, create, iptcap) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating masq rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* snat rules */
							else if(create->ruletype == RT_SNAT)
							{
								if(create_rule_snat(debuglvl, ruleset, rule, create, iptcap) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating snat rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* portforward rules */
							else if(create->ruletype == RT_PORTFW)
							{
								if(create_rule_portfw(debuglvl, ruleset, rule, create, iptcap) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating portfw rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							else if(create->ruletype == RT_REDIRECT)
							{
								if(create_rule_redirect(debuglvl, ruleset, rule, create, iptcap) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating redirect rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							else
							{
								(void)vrprint.error(-1, "Internal Error", "unknown ruletype '%d' (in: %s).", create->ruletype, __FUNCTION__);
								return(-1);
							}
						} /* end createloop */

						/* get the next service/portrange */
						if(create->service->PortrangeList.len > 0)
						{
							if(d_list_node_is_bot(debuglvl, port_d_node))
							{
								/* rewind */
								port_d_node = create->service->PortrangeList.top;
							}
							else
							{
								/* next */
								port_d_node = port_d_node->next;
							}

							/* listport option */
							if(create->option.listenport == 1)
							{
								if(listenport_d_node->next != NULL)
									listenport_d_node = listenport_d_node->next;
							}

							/* remoteport  option */
							if(create->option.remoteport == 1)
							{
								if(remoteport_d_node->next != NULL)
									remoteport_d_node = remoteport_d_node->next;
							}
						}
					}
				} /* end if active */
				else
				{
					if(debuglvl >= LOW)
						(void)vrprint.debug("Debug", "%s: rule not created, not active.", __FUNCTION__);
				}

				/* set the next group member TO GROUP */
				if(create->to->type == TYPE_GROUP)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "%s: TO: set the next (or prev) group member.", __FUNCTION__);

					/* get the new to_d_node

					   if we have reached the bot of the list, start all over again
					 */
					if(!d_list_node_is_bot(debuglvl, to_d_node))
						to_d_node = to_d_node->next;
					else
						to_d_node = create->to->GroupList.top;

					if(to_d_node)
					{
						if(!(to_member_ptr = to_d_node->data))
						{
							(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
							return(-1);
						}

						if(to_member_ptr->active == 0)
						{
							if(debuglvl >= MEDIUM)
								(void)vrprint.debug("Debug", "%s: group member '%s' is inactive", __FUNCTION__, to_member_ptr->name);

							to_active = 0;
						}
						else
						{
							if(debuglvl >= MEDIUM)
								(void)vrprint.debug("Debug", "%s: group member '%s' is active", __FUNCTION__, to_member_ptr->name);

							to_active = 1;
						}
					}
				}
			}

			/* set the next group member of the FROM GROUP */
			if(create->from->type == TYPE_GROUP)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: FROM: set the next (or prev) group member.", __FUNCTION__);
				
				/* get the new from_d_node

				   if we have reached the bot of the list, start all over again
				 */
				if(!d_list_node_is_bot(debuglvl, from_d_node))
					from_d_node = from_d_node->next;
				else
					from_d_node = create->from->GroupList.top;

				if(from_d_node)
				{
					if(!(from_member_ptr = from_d_node->data))
					{
						(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
						return(-1);
					}

					if(from_member_ptr->active == 0)
					{
						if(debuglvl >= MEDIUM)
							(void)vrprint.debug("Debug", "%s: group member '%s' is inactive", __FUNCTION__, from_member_ptr->name);

						from_active = 0;
					}
					else
					{
						if(debuglvl >= MEDIUM)
							(void)vrprint.debug("Debug", "%s: group member '%s' is active", __FUNCTION__, from_member_ptr->name);

						from_active = 1;
					}
				}
			}
		}
		/* not creating because of 'int=' option in the rule */
		else
		{
			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNCTION__, "don't create because of interface option '%s'.", create->option.interface);
		}
//
//// END LOOP INTERFACES
//
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug("Debug", "%s: rules created: (i: %2d, o: %2d, f: %2d, e: %2d, p: %2d)", __FUNCTION__, create->iptcount.input, create->iptcount.output, create->iptcount.forward, create->iptcount.preroute, create->iptcount.postroute);
	} //end if protect

	/* free the temp data */
	free(rule);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** end **, return=%d", __FUNCTION__, retval);
	
	return(retval);
}


/*	remove_rule

	Function for removing iptables rules.

	Returncodes:
		 0: ok
		-1: error
*/
int
remove_rule(const int debuglvl, int chaintype, int first_ipt_rule, int rules)
{
	int	retval=0,
		i;
	char	cmd[MAX_PIPE_COMMAND];
	char	chain[sizeof(RuleCache.chain)];

	if(debuglvl >= HIGH)
	{
		(void)vrprint.debug("Debug", "remove_rule: ** start **");
		(void)vrprint.debug("Debug", "chain: %d, ipt: %d, rules: %d", chaintype, first_ipt_rule, rules);
	}

	/* determine from which chain to delete */
	if(chaintype == RT_INPUT)
	{
		strlcpy(chain, "-D INPUT", sizeof(chain));
	}
	else if(chaintype == RT_OUTPUT)
	{
		strlcpy(chain, "-D OUTPUT", sizeof(chain));
	}
	else if(chaintype == RT_FORWARD)
	{
		strlcpy(chain, "-D FORWARD", sizeof(chain));
	}
	else if(chaintype == RT_PORTFW || chaintype == RT_REDIRECT)
	{
		strlcpy(chain, "-t nat -D PREROUTING", sizeof(chain));
	}
	else if(chaintype == RT_MASQ || chaintype == RT_SNAT)
	{
		strlcpy(chain, "-t nat -D POSTROUTING", sizeof(chain));
	}
	else
	{
		(void)vrprint.error(-1, "Error", "unknown chaintype %d (remove_rule).", chaintype);
		return(-1);
	}

	for(i = 0; i < rules; i++)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "remove_rule: cmd: %s %s %d", conf.iptables_location, chain, first_ipt_rule);

		snprintf(cmd, sizeof(cmd), "%s %s %d", conf.iptables_location, chain, first_ipt_rule);
		if(pipe_command(debuglvl, cmd, PIPE_VERBOSE) != 0)
		{
			(void)vrprint.error(-1, "Error", "remove_rule: pipe error. This command failed: '%s'.", cmd);
			return(-1);
		}
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "remove_rule: ** end **, return=%d", retval);

	return(retval);
}


int
create_interface_protect_rules(const int debuglvl, Interfaces *interfaces)
{
	struct RuleCache_	*create = NULL;
	d_list_node		*d_node = NULL,
				*if_d_node = NULL;
	struct RuleData_	*rule_ptr = NULL;
	struct InterfaceData_	*iface_ptr = NULL;


	/* safety */
	if(!interfaces)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(bash_out == 1)
		fprintf(stdout, "\n# Loading interfaces protection rules...\n");

	/* loop through the interfaces */
	for(d_node = interfaces->list.top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		/* now loop through the ruleslist */
		for(if_d_node = iface_ptr->ProtectList.top; if_d_node; if_d_node = if_d_node->next)
		{
			if(!(rule_ptr = if_d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}
			if(!(create = &rule_ptr->rulecache))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}
			if(!create->who_int)
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			/* here we print the description if we are in bashmode */
			if(bash_out == 1 && create->description != NULL)
			{
				fprintf(stdout, "\n# %s\n", create->description);
		
				free(create->description);
				create->description = NULL;
			}

			/*
				prot rule, proc only for interface
			*/
			if(create->danger.solution == PROT_PROC_INT)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "protect proc (int)... ");

				/*	only set if int is active and up.
		
					NOTE: this is different from iptables rules, which are also created if the interface is
					down. However the proc entries are only available when the interface is up!
				*/
				if(create->who_int->active && create->who_int->up)
				{
					if(debuglvl >= MEDIUM)
						(void)vrprint.debug("Debug", "Setting '%d' to '%s' (where * is %s)... ", create->danger.proc_set_on, create->danger.proc_entry, create->who_int->device);

					if(set_proc_entry(debuglvl, create->danger.proc_entry, create->danger.proc_set_on, create->who_int->device) != 0)
					{
						/* if it fails, we dont really care, its not fatal */
						(void)vrprint.error(-1, "Error", "setting proc entry failed (in: %s:%d).", __FUNCTION__, __LINE__);
					}
				}
				else
				{
					if(bash_out == 1)
					{
						fprintf(stdout, "# not created: interface is inactive or down.\n");
					}
				}
			}
			/* Whoops, this is serious. */
			else
			{
				(void)vrprint.error(-1, "Internal Error", "unknown protect danger type %d (in: %s:%d).", create->danger.solution, __FUNCTION__, __LINE__);
				return(-1);
			}
		}
	}

	return(0);
}


int
create_system_protectrules(const int debuglvl, struct vuurmuur_config *conf)
{
	int	result = 0;

	/* safety */
	if(!conf)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "protect proc systemwide... ");


	/* syncookies */
	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "Setting '%d' to '%s'... ", conf->protect_syncookie, "/proc/sys/net/ipv4/tcp_syncookies");

	result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/tcp_syncookies", conf->protect_syncookie, NULL);
	if(result != 0)
	{
		/* if it fails, we dont really care, its not fatal */
		(void)vrprint.error(-1, "Error", "set_proc_entry failed (in: create_rule, prot_proc_sys).");
	}


	/* echo broadcasts */
	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "Setting '%d' to '%s'... ", conf->protect_echobroadcast, "/proc/sys/net/ipv4/icmp_echo_ignore_broadcasts");

	result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/icmp_echo_ignore_broadcasts", conf->protect_echobroadcast, NULL);
	if(result != 0)
	{
		/* if it fails, we dont really care, its not fatal */
		(void)vrprint.error(-1, "Error", "set_proc_entry failed (in: create_rule, prot_proc_sys).");
	}

	return(0);
}


int
create_normal_rules(const int debuglvl, /*@null@*/RuleSet *ruleset, Rules *rules, IptCap *iptcap, char *forward_rules)
{
	d_list_node		*d_node = NULL;
	struct RuleData_	*rule_ptr = NULL;
	char			active = 0;
	int			rulescount = 0;


	/* safety */
	if(!rules || !forward_rules)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* walk trough the ruleslist and create the rules */
	for(d_node = rules->list.top; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		/* asume active when we begin */
		active = 1;

		/* count the rules */
		rulescount++;

		/* check normal rule */
		if(rule_ptr->rulecache.from && rule_ptr->rulecache.to && rule_ptr->rulecache.service)
		{
			if(!rule_ptr->rulecache.from->active || !rule_ptr->rulecache.to->active || !rule_ptr->rulecache.service->active)
			{
				/* if one of above is not active, set active to NO */
				active = 0;
			}
		}

		/* check protect rule */
		if(rule_ptr->rulecache.who)
		{
			if(rule_ptr->rulecache.who->active == 0)
			{
				active = 0;
			}
		}

		/* create the rule */
		if(active)
		{
			if(rule_ptr->rulecache.from && rule_ptr->rulecache.to && rule_ptr->rulecache.service)
			{
				if(create_rule(debuglvl, ruleset, &rule_ptr->rulecache, iptcap) == 0)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "rule created succesfully.");

					if(rule_ptr->rulecache.iptcount.forward > 0)
						*forward_rules = 1;
				}
				else
				{
					(void)vrprint.warning("Warning", "Creating rule %d failed.", rulescount);
				}
			}
			else
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}
		}
		else
		{
			(void)vrprint.info("Note", "Rule %d not created: inactive.", rulescount);
		}
	}

	return(0);
}


/*	clear_all_iptables_rule

	Clears all rules and chains created by Vuurmuur.
	For use with the -C commandline option.

	Returncodes:
		-1: error
		 0: ok
*/
int
clear_all_iptables_rules(const int debuglvl)
{
	int	retval = 0,
		result = 0;
	char	cmd[MAX_PIPE_COMMAND] = "";

	/* flush everything */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
	if(result < 0)
		retval = -1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
	if(result < 0)
		retval = -1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
	if(result < 0)
		retval = -1;

	/* flush anti spoof old rules */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush ANTISPOOF 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	/* flush the BLOCKLIST */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush BLOCKLIST 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	/* flush the BLOCKLIST */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush BLOCK 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	/* flush the synlimit chain */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush SYNLIMIT 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);

	/* delete the chains */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X ANTISPOOF 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X SYNLIMIT 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X BLOCKLIST 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X BLOCK 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, PIPE_QUIET);

	/* set default polices to ACCEPT */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy INPUT ACCEPT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
	if(result < 0)
		retval=-1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy OUTPUT ACCEPT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
	if(result < 0)
		retval=-1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy FORWARD ACCEPT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, PIPE_VERBOSE);
	if(result < 0)
		retval=-1;

	return(retval);
}
